import os, sys
import sphinx_rtd_theme as theme
sys.path.insert(0, os.path.dirname(__file__))
source_suffix = '.rst'
master_doc = 'index'
project = u'git-spindle'
copyright = u'2012-2020, Dennis Kaarsemaker'
version = '4.0'
release = '4.0.1'
exclude_patterns = ['_build']
pygments_style = 'sphinx'
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
html_theme_options = {
}
html_theme_path = [theme.get_html_theme_path()]
html_show_sourcelink = False
html_show_sphinx = False
extensions = ['ansicolor']
man_pages = [
    ('github', 'git-hub', 'GitHub integration', 'Dennis Kaarsemaker', '1'),
    ('gitlab', 'git-lab', 'GitLab integration', 'Dennis Kaarsemaker', '1'),
    ('bitbucket', 'git-bb', 'BitBucket integration', 'Dennis Kaarsemaker', '1'),
    ('bitbucket', 'git-bucket', 'BitBucket integration', 'Dennis Kaarsemaker', '1'),
]



# Set the spindle attribute
import github3.gists

# Monkeypatch github3.gists.Gist to behave more like a repo
github3.gists.gist._Gist.ssh_url = property(lambda self: 'git@gist.github.com:/%s.git' % self.id)
github3.gists.gist._Gist.clone_url = property(lambda self: self.git_pull_url)
github3.gists.gist._Gist.git_url = property(lambda self: 'git://gist.github.com/%s.git' % self.id)
github3.gists.gist._Gist.name = property(lambda self: self.id)
github3.gists.gist._Gist.private = property(lambda self: not self.public)
github3.gists.gist._Gist.create_fork = github3.gists.Gist.fork
# XXX - There is nothing in the API output that indicates forkedness
github3.gists.gist._Gist.fork = False
github3.gists.gist._Gist.issues = lambda self, *args, **kwargs: []
def _gist_events(self, number=300):
    for event in self.history[:number]:
        yield GistEvent(event, self)
class GistEvent(object):
    type = 'GistHistoryEvent'
    def __init__(self, history, gist):
        self.created_at = history.committed_at
        self.additions = history.additions
        self.deletions = history.deletions
        self.actor = history.user
        if not self.actor.login:
            self.actor = gist.owner
        self.repo = ('gist', gist.name)
github3.gists.Gist.events = _gist_events
class Content(object):
    def __init__(self, file):
        self.decoded = file.content()
def _gist_contents(self, path, ref):
    # XXX ignore ref for now, can't do much with it
    for f in self.files():
        if f.filename == path:
            return Content(f)
github3.gists.Gist.file_contents = _gist_contents

# This is needed when working with the device authentication flow, as we're
# getting fewer attributes back
import github3.users
github3.users.AuthenticatedUser._update_attributes = github3.users.User._update_attributes

# Monkeypatch docopt to support our git-clone-options-hack
import docopt
known_options = {
    'clone': (
        '-q', '--quiet', '-v', '--verbose', '-n', '--no-checkout', '--bare',
        '--mirror', '--reference=<repository>', '--progress', '-o <oname>', '--origin=<oname>',
        '-b <name>', '--branch=<name>', '-u <upload-pack>', '--upload-pack=<upload-pack>',
        '--template=<template-directory>', '-c <key-value>', '--config=<key-value>',
        '--depth=<depth>', '--single-branch', '--no-single-branch', '--recursive', '--recurse-submodules',
        '--separate-git-dir=<git_dir>'),
}

class GitOption(docopt.Option):
    def match(self, left, collected=None):
        if left and isinstance(left[0], docopt.Option) and ((self.short == left[0].short and self.short) or (self.long == left[0].long and self.long)):
            opt = left.pop(0)
            # Hijack the command argument to store the data
            cmd = [x for x in collected if type(x).__name__ == 'Command'][1]
            if not isinstance(cmd.value, list):
                cmd.value = []
            cmd.value.append(opt.short or opt.long)
            if opt.argcount:
                cmd.value.append(opt.value)
        return True, left, collected

    def __repr__(self):
        return 'Git' + super(GitOption, self).__repr__()


def parse_atom(tokens, options):
    if len(tokens) > 3 and (tokens[0], tokens[2]) == ('[', ']') and \
        tokens[1].startswith('git-') and tokens[1].endswith('-options'):
            token = tokens[1][4:-8]
            tokens.move(); tokens.move(); tokens.move()
            ret = []
            for opt in known_options[token]:
                opt = docopt.parse_pattern(opt, []).children[0]
                opt = GitOption(opt.short, opt.long, opt.argcount, opt.value)
                ret.append(opt)
                options.append(opt)
            return ret
    return docopt.orig_parse_atom(tokens, options)

docopt.orig_parse_atom = docopt.parse_atom
docopt.parse_atom = parse_atom

def formal_usage(printable_usage):
    usage = real_printable_usage(printable_usage).splitlines()
    ret = []
    for num, line in enumerate(usage):
        if line[0].isupper() and usage[num+1].startswith('  git'):
            continue
        ret.append(line)
    return docopt.orig_formal_usage('\n'.join(ret))
real_printable_usage = docopt.printable_usage
docopt.printable_usage = lambda x: x
docopt.orig_formal_usage = docopt.formal_usage
docopt.formal_usage = formal_usage



from gitspindle import *
from gitspindle.ansi import *
import datetime
import getpass
import github3
import github3.gists
import github3.session
import glob
import os
import re
import requests
import socket
import sys
import tempfile
import time
import webbrowser

class RateLimitedSession(github3.session.GitHubSession):
    """A session subclass that warns when we're approaching API rate limits"""
    def __init__(self, *args, **kwargs):
        super(RateLimitedSession, self).__init__(*args, **kwargs)
        self.warned = False

    def request(self, *args, **kwargs):
        response = super(RateLimitedSession, self).request(*args, **kwargs)

        limit = int(response.headers.get('x-ratelimit-limit', 0))
        remaining = int(response.headers.get('x-ratelimit-remaining', 0))
        reset = int(response.headers.get('x-ratelimit-reset', 0))

        if limit and (remaining < 0.20 * limit) and not self.warned:
            msg = "You are approaching the API rate limit. Only %d/%d requests remain until %s"
            msg = msg % (remaining, limit, time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(reset)))
            print(wrap(msg, fgcolor.red, attr.bright))
            self.warned = True
        return response

class GitHub(GitSpindle):
    prog = 'git hub'
    what = 'GitHub'
    spindle = 'github'
    hosts = ['github.com', 'www.github.com', 'gist.github.com']
    api = github3
    client_id = 'Iv1.6b6ddacfe13b6331'

    # Support functions
    def login(self):
        host = self.config('host')
        if host and host not in ('https://api.github.com', 'api.github.com'):
            if not host.startswith(('http://', 'https://')):
                try:
                    requests.get('https://' + host)
                except:
                    err("%s is not reachable via https. Use http://%s to use the insecure http protocol" % (host, host))
                host = 'https://' + host
            self.gh = github3.GitHubEnterprise(url=host, session=RateLimitedSession())
        else:
            self.gh = github3.GitHub(session=RateLimitedSession())

        user = self.config('user')
        if not user:
            user = input("GitHub user: ").strip()
            self.config('user', user)

        token = self.config('token')
        if not token:
            url = self.gh._build_url('login/device/code').replace('api.github.com','github.com')
            resp = requests.post(url,data={'client_id': self.client_id, 'scope': 'user,repo,gist,admin:public_key,admin:repo_hook,admin:org'}, headers={'Accept': 'application/json'})
            if resp.status_code != 200:
                # We don't support the device flow yet
                print("Generate a personal access token at %s and enter it below" % self.gh._build_url('settings/tokens').replace('api.github.com','github.com'))
                print("The required scopes are: user, repo, gist, admin:public_key, admin:repo_hook, admin:org")
                token = getpass.getpass("GitHub access token: ").strip()
            else:
                data = resp.json()
                print("Please visit %s and enter the following one-time code: %s" % (data['verification_uri'], data['user_code']))
                while True:
                    url = self.gh._build_url('login/oauth/access_token').replace('api.github.com','github.com')
                    time.sleep(data['interval'])
                    resp = requests.post(url, data={'client_id': self.client_id, 'device_code': data['device_code'], 'grant_type': 'urn:ietf:params:oauth:grant-type:device_code'}, headers={'Accept': 'application/json'})
                    if resp.status_code != 200:
                        err("Authentication failed, http code %s" % resp.status_code)
                    data2 = resp.json()
                    error = data2.get("error", None)
                    if error:
                        if error == 'authorization_pending':
                            continue
                        if error == 'slow_down':
                            data['interval'] = data2['interval']
                            continue
                        if error == 'access_denied':
                            err("Authentication canceled in the web browser")
                        if error == 'expired_token':
                            err("Authentication timed out")
                        err("Internal authentication error")
                    token = data2['access_token']
                    break
            self.config('token', token)
            location = '%s - do not share this file' % self.config_file
            if self.use_credential_helper:
                location = 'git\'s credential helper'
            print("A GitHub authentication token is now stored in %s" % location)
            print("To revoke access, visit https://github.com/settings/applications")

        if not user or not token:
            err("No user or token specified")
        self.gh.login(username=user, token=token)
        try:
            self.me = self.gh.me()
            self.my_login = self.me.login
        except github3.GitHubError:
            raise
            # Token obsolete
            self.config('token', None)
            self.login()

    def parse_url(self, url):
        if url.hostname == 'gist.github.com':
            return ['gist', url.path.split('/')[-1]]
        else:
            return ([self.my_login] + url.path.split('/'))[-2:]

    def get_repo(self, remote, user, repo):
        if user == 'gist':
            # This is a gist, not a normal repo
            repo_ = self.gh.gist(repo)
            if not repo_:
                err("Gist %s does not exist" % repo)
        else:
            repo_ = self.gh.repository(user, repo)

        return repo_

    def parent_repo(self, repo):
        if repo.fork:
            # In search results or lists parent info is not returned with a repository
            return getattr(repo, 'parent', None) or self.gh.repository(repo.owner.login, repo.name).parent

    def clone_url(self, repo, opts):
        if opts['--ssh'] or repo.private:
            return repo.ssh_url
        if opts['--http']:
            return repo.clone_url
        if opts['--git']:
            return repo.git_url
        if self.my_login == repo.owner.login:
            return repo.ssh_url
        return repo.clone_url

    def api_root(self):
        if hasattr(self, 'gh'):
            return self.gh.session.base_url
        host = self.config('host')
        if not host:
            return 'https://api.github.com'
        return host.rstrip('/') + '/api/v3'

    def find_template(self, repo, template):
        template = template.lower()
        contents = None
        for dir in ('/', '/.github/'):
            try:
                files = repo.directory_contents(dir)
            except github3.exceptions.NotFoundError:
                continue
            files = dict([(k.lower(), v) for k, v in files])

            if template in files:
                contents = files[template]
            else:
                for file in files:
                    if file.startswith(template + '.'):
                        contents = files[file]
        if contents:
            contents = contents.name, self.gh.session.get(contents.download_url, stream=True).text
        return contents

    # Commands

    @command
    def add_collaborator(self, opts):
        """<user>...
           Add a user as collaborator"""
        repo = self.repository(opts)
        for user in opts['<user>']:
            repo.add_collaborator(user)

    @command
    def add_deploy_key(self, opts):
        """[--read-only] <key>...
           Add a deploy key"""
        repo = self.repository(opts)
        url = repo._build_url('keys', base_url=repo._api)
        for arg in opts['<key>']:
            with open(arg) as fd:
                algo, key, title = fd.read().strip().split(None, 2)
            key = "%s %s" % (algo, key)
            print("Adding deploy key %s" % arg)
            repo.create_key(title=title, key=key, read_only=opts['--read-only'])

    @command
    def add_hook(self, opts):
        """<name> [<setting>...]
           Add a repository hook"""
        repo = self.repository(opts)
        for hook in repo.hooks():
            if hook.name == opts['<name>']:
                err("Hook %s already exists" % opts['<name>'])
        settings = dict([x.split('=', 1) for x in opts['<setting>']])
        for key in settings:
            if settings[key].isdigit():
                settings[key] = int(settings[key])
        events = settings.pop('events', 'push').split(',')
        repo.create_hook(opts['<name>'], settings, events)

    @command
    def apply_pr(self, opts):
        """[--ssh|--http|--git] <pr-number>
           Applies a pull request as a series of cherry-picks"""
        repo = self.repository(opts)
        pr = repo.pull_request(opts['<pr-number>'])
        if not pr:
            err("Pull request %s does not exist" % opts['<pr-number>'])
        print("Applying PR#%d from %s: %s" % (pr.number, pr.user.refresh().name or pr.user.login, pr.title))
        # Warnings
        warned = False
        cbr = self.gitm('rev-parse', '--symbolic-full-name', 'HEAD').stdout.strip().replace('refs/heads/','')
        if cbr != pr.base.ref:
            print(wrap("Pull request was filed against %s, but you're on the %s branch" % (pr.base.ref, cbr), fgcolor.red))
            warned = True
        if pr.merged_at:
            print(wrap("Pull request was already merged at %s by %s" % (pr.merged_at, pr.merged_by), fgcolor.red))
        if not pr.mergeable or pr.mergeable_state != 'clean':
            print(wrap("Pull request will not apply cleanly", fgcolor.red))
            warned = True
        if pr.state == 'closed':
            print(wrap("Pull request has already been closed", fgcolor.red))
            warned = True
        if warned:
            if not self.question("Continue?", default=False):
                sys.exit(1)
        # Fetch PR if needed
        sha = self.git('rev-parse', '--verify', 'refs/pull/%d/head' % pr.number).stdout.strip()
        if sha != pr.head.sha:
            print("Fetching pull request")
            url = self.clone_url(self.gh.repository(pr.repository[0].replace('repos/', ''), pr.repository[1]), opts)
            self.gitm('fetch', url, 'refs/pull/%d/head:refs/pull/%d/head' % (pr.number, pr.number), redirect=False)
        head_sha = self.gitm('rev-parse', 'HEAD').stdout.strip()
        if self.git('merge-base', pr.head.sha, head_sha).stdout.strip() == head_sha:
            print("Fast-forward merging %d commit(s): %s..refs/pull/%d/head" % (pr.commits_count, pr.base.ref, pr.number))
            self.gitm('merge', '--ff-only', 'refs/pull/%d/head' % pr.number, redirect=False)
        else:
            print("Cherry-picking %d commit(s): %s..refs/pull/%d/head" % (pr.commits_count, pr.base.ref, pr.number))
            self.gitm('cherry-pick', '%s..refs/pull/%d/head' % (pr.base.ref, pr.number), redirect=False)

    @command
    @wants_parent
    def add_remote(self, opts):
        """[--ssh|--http|--git] <user> [<name>]
           Add user's fork as a named remote. The name defaults to the user's loginname"""
        for fork in self.repository(opts).forks():
            if fork.owner.login in opts['<user>']:
                url = self.clone_url(fork, opts)
                name = opts['<name>'] or fork.owner.login
                self.gitm('remote', 'add', '-f', name, url, redirect=False)

    @command
    def add_public_keys(self, opts):
        """[<key>...]
           Adds keys to your public keys"""
        if not opts['<key>']:
            opts['<key>'] = glob.glob(os.path.join(os.path.expanduser('~'), '.ssh', 'id_*.pub'))
        existing = [x.key for x in self.gh.keys()]
        for arg in opts['<key>']:
            with open(arg) as fd:
                algo, key, title = fd.read().strip().split(None, 2)
            key = "%s %s" % (algo, key)
            if key in existing:
                continue
            print("Adding %s" % arg)
            self.gh.create_key(title=title, key=key)

    @command
    def browse(self, opts):
        """[--parent] [<repo>] [<section>]
           Open the GitHub page for a repository in a browser"""
        sections = ['issues', 'pulls', 'wiki', 'branches', 'releases', 'contributors', 'graphs', 'settings']
        if opts['<repo>'] in sections and not opts['<section>']:
            opts['<repo>'], opts['<section>'] = None, opts['<repo>']
        repo = self.repository(opts)
        url = repo.html_url
        if opts['<section>']:
            url += '/' + opts['<section>']
        webbrowser.open_new_tab(url)

    @command
    def calendar(self, opts):
        """[<user>]
           Show a timeline of a user's activity"""
        user = (opts['<user>'] or [self.my_login])[0]
        months = []
        rows = [[],[],[],[],[],[],[]]
        commits = []

        data = requests.get('https://github.com/users/%s/contributions' % user).text
        # Sorry, zalgo!
        data = re.findall(r'data-count="(.*?)" data-date="(.*?)"', data)
        y, m, d = [int(x) for x in data[0][1].split('-')]
        wd = (datetime.date(y,m,d).weekday()+1) % 7
        for i in range(wd):
            rows[i].append((None,None))
        if wd:
            months.append(m)
        for (count, date) in data:
            count = int(count)
            y, m, d = [int(x) for x in date.split('-')]
            wd = (datetime.date(y,m,d).weekday()+1) % 7
            rows[wd].append((d, count))
            if not wd:
                months.append(m)
            if count:
                commits.append(count)

        # Print months
        sys.stdout.write("  ")
        last = -1
        skip = months[2] != months[0]
        monthtext = ('', 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')
        for month in months:
            if month != last:
                sys.stdout.write(monthtext[month] + ' ')
                skip = True
                last = month
            elif not skip:
                sys.stdout.write('  ')
            else:
                skip = False
        print("")

        # Print commits
        days = 'SMTWTFS'
        commits.sort()
        if len(commits) < 2:
            p5 = p15 = p35 = 0
        else:
            p5  = commits[min(int(round(len(commits) * 0.95)), len(commits)-1)]
            p15 = commits[min(int(round(len(commits) * 0.85)), len(commits)-1)]
            p35 = commits[min(int(round(len(commits) * 0.65)), len(commits)-1)]
        blob1 = b'\xe2\x96\xa0'.decode('utf-8')
        blob2 = b'\xe2\x97\xbc'.decode('utf-8')
        for rnum, row in enumerate(rows):
            if rnum % 2:
                sys.stdout.write(days[rnum] + " ")
            else:
                sys.stdout.write("  ")
            for (day, count) in row:
                if count is None:
                    color = attr.conceal
                elif count > p5:
                    color = fgcolor.xterm(22)
                elif count > p15:
                    color = fgcolor.xterm(28)
                elif count > p35:
                    color = fgcolor.xterm(64)
                elif count:
                    color = fgcolor.xterm(65)
                else:
                    color = fgcolor.xterm(237)
                if day == 1:
                    msg = wrap(blob2, attr.underline, color)
                    sys.stdout.write(msg)
                else:
                    msg = wrap(blob1, color)
                    sys.stdout.write(msg)
                sys.stdout.write(' ')
            print("")

    @command
    def cat(self, opts):
        """<file>...
           Display the contents of a file on GitHub"""
        for arg in opts['<file>']:
            repo, ref, file = ([None, None] + arg.split(':',2))[-3:]
            user = None
            if repo:
                user, repo = ([None] + repo.split('/'))[-2:]
                repo = self.gh.repository(user or self.my_login, repo)
            else:
                repo = self.repository(opts)
                file = self.rel2root(file)
            content = repo.file_contents(file, ref=ref or repo.default_branch)
            if not content:
                err("No such file: %s" % arg)
            if content.type != 'file':
                err("Not a regular file: %s" % arg)
            resp = self.gh.session.get(content.download_url, stream=True)
            for chunk in resp.iter_content(4096):
                os.write(sys.stdout.fileno(), chunk)

    @command
    def check_pages(self, opts):
        """[<repo>] [--parent]
           Check the github pages configuration and content of your repo"""
        repo = self.repository(opts)

        if opts['<repo>']:
            self.clone(opts)
            os.chdir(repo.name)

        def warning(msg, url=None):
            print(wrap(msg, fgcolor.yellow))
            if url:
                print(wrap(url, attr.faint))

        def error(msg, url=None):
            print(wrap(msg, fgcolor.red))
            if url:
                print(wrap(url, attr.faint))

        # Old-style $user.github.com repos
        if repo.name.lower() == repo.owner.login.lower() + '.github.com':
            warning("Your repository is named %s.github.com, but should be named %s.github.io" % (repo.owner.login, repo.owner.login),
                    "https://help.github.com/articles/user-organization-and-project-pages/#user--organization-pages")

#        if repo.name.lower() == repo.owner.login.lower() + '.github.io' and repo.name != repo.name.lower():
#            error("You should not have capital letters in your repository name, please rename it from %s to %s" % (repo.name, repo.name.lower()))

        # Which branch do we check?
        if repo.name.lower() in (repo.owner.login.lower() + '.github.com', repo.owner.login.lower() + '.github.io'):
            branchname = 'master'
        else:
            branchname = 'gh-pages'

        # Do we have local changes?
        if self.git('rev-parse', '--symbolic-full-name', 'HEAD').stdout.strip() == 'refs/heads/%s' % branchname and self.git('status', '--porcelain').stdout.strip():
            warning("You have uncommitted changes. This tool checks the latest commit, not the working tree")

        # Do we have a pages branch?
        local = remote_tracking = remote = None

        output = self.git('ls-remote', repo.remote or 'origin', 'refs/heads/%s' % branchname).stdout
        for line in output.splitlines():
            remote = line.split()[0]
        if not remote:
            error("You have no %s branch on GitHub" % branchname,
                  "https://help.github.com/articles/user-organization-and-project-pages/")

        output = self.git('for-each-ref', '--format=%(refname) %(objectname) %(upstream:trackshort)',
                          'refs/remotes/%s/%s' % (repo.remote or 'origin', branchname),
                          'refs/heads/%s' % branchname,).stdout
        for line in output.splitlines():
            if line.startswith('refs/heads'):
                ref, sha, ahead = line.split()
                local = sha
                if ahead == '<':
                    warning("Your local %s branch is behind the one on GitHub" % branchname)
                elif ahead == '>':
                    warning("Your local %s branch is ahead of the one on GitHub" % branchname)
                elif ahead == '<>':
                    warning("Your local %s branch has diverged from the one on GitHub" % branchname)
            elif line.startswith('refs/remotes'):
                ref, sha = line.split()
                remote_tracking = sha
                if remote != remote_tracking:
                    warning("You need to fetch %s from GitHub to get its latest revision" % branchname)

        if not local or not remote_tracking:
            warning("You have no %s branch locally" % branchname,
                    "https://help.github.com/articles/user-organization-and-project-pages/")

        if local:
            ref = 'refs/heads/%s' % branchname
        elif remote_tracking:
            ref = 'refs/remotes/%s/%s' % (repo.remote, branchname)
        files = self.git('ls-tree', '-r', '--name-only', ref).stdout.splitlines()

        # Do we have an index.html
        if 'index.html' not in files:
            warning("You have no index.html")

        # Do we need .nojekyll (dirs starting with underscores)
        if '.nojekyll' not in files and '_config.yml' not in files:
            for file in files:
                if file.startswith('_'):
                    warning("You have filenames starting with underscores, but no .nojekyll file",
                            "https://help.github.com/articles/using-jekyll-with-pages/#turning-jekyll-off")
                break

        # Do we have unverified emails
        if repo.owner.login == self.me.login:
            for mail in self.gh.emails():
                if not mail.verified:
                    error("Unverified %s email address: %s" % (mail.primary and 'primary' or 'secondary', mail.email))

        # Do we have a custom CNAME. Check DNS (Use meta api for A records)
        for file in files:
            if file.lower() == 'cname':
                if file != 'CNAME':
                    error("The CNAME file must be named in all caps",
                          "https://help.github.com/articles/adding-a-cname-file-to-your-repository/")
                cname = self.git('--no-pager', 'show', '%s:%s' % (ref, file)).stdout.strip()
                pages_ips = self.gh.meta()['pages']
                try:
                    import publicsuffix
                except ImportError:
                    import gitspindle.public_suffix as publicsuffix
                expect_cname = publicsuffix.PublicSuffixList(publicsuffix.fetch()).get_public_suffix(cname) != cname
                try:
                    import dns
                    import dns.resolver
                    resolver = dns.resolver.Resolver()
                    answer = resolver.query(cname)
                    for rrset in answer.response.answer:
                        name = rrset.name.to_text().rstrip('.')
                        if name == cname:
                            for rr in rrset:
                                if rr.rdtype == dns.rdatatype.A and expect_cname:
                                    warning("You should use a CNAME record for non-apex domains",
                                            "https://help.github.com/articles/tips-for-configuring-a-cname-record-with-your-dns-provider/")
                                if rr.rdtype == dns.rdatatype.A and rr.address not in pages_ips:
                                    error("IP address %s is incorreect for a pages site, use only %s" % (rr.address, ', '.join(pages_ips)),
                                          "https://help.github.com/articles/tips-for-configuring-a-cname-record-with-your-dns-provider/")
                                if rr.rdtype == dns.rdatatype.CNAME and rr.target != '%s.github.io.' % repo.owner.login:
                                    error("CNAME %s -> %s is incorrect, should be %s -> %s" % (name, rr.target, name, '%s.github.io.' % repo.owner.login),
                                          "https://help.github.com/articles/tips-for-configuring-an-a-record-with-your-dns-provider/")
                except ImportError:
                    if hasattr(self.shell, 'dig'):
                        lines = self.shell.dig('+nocomment', '+nocmd', '+nostats', '+noquestion', cname).stdout.splitlines()
                        for line in lines:
                            rname, ttl, _, rtype, value = line.split(None, 4)
                            if rname.rstrip('.') == cname:
                                if rtype == 'A' and expect_cname:
                                    warning("You should use a CNAME record for non-apex domains",
                                            "https://help.github.com/articles/tips-for-configuring-a-cname-record-with-your-dns-provider/")
                                if rtype == 'A' and value not in pages_ips:
                                    error("IP address %s is incorreect for a pages site, use only %s" % (value, ', '.join(pages_ips)),
                                          "https://help.github.com/articles/tips-for-configuring-a-cname-record-with-your-dns-provider/")
                                if rtype == 'CNAME' and value != '%s.github.io.' % repo.owner.login:
                                    error("CNAME %s -> %s is incorrect, should be %s -> %s" % (rname, value, rname, '%s.github.io.' % repo.owner.login),
                                          "https://help.github.com/articles/tips-for-configuring-an-a-record-with-your-dns-provider/")
                    else:
                        error("Cannot check DNS settings. Please install dnspython or dig")
                break

    @command
    def clone(self, opts, repo=None):
        """[--ssh|--http|--git] [--triangular [--upstream-branch=<branch>]] [--parent] [git-clone-options] <repo> [<dir>]
           Clone a repository by name"""
        if not repo:
            repo = self.repository(opts)
        url = self.clone_url(repo, opts)
        args = opts['extra-opts']
        args.append(url)
        dir = opts['<dir>'] and opts['<dir>'][0] or repo.name
        if '--bare' in args:
            dir += '.git'
        args.append(dir)

        self.gitm('clone', *args, redirect=False).returncode
        if repo.fork:
            os.chdir(dir)
            self.set_origin(opts, repo=repo)

    @command
    def collaborators(self, opts):
        """[<repo>]
           List collaborators of a repository"""
        repo = self.repository(opts)
        users = list(repo.collaborators())
        users.sort(key = lambda user: user.login)
        for user in users:
            print(user.login)

    @command
    def create(self, opts):
        """[--private] [--org=<org>] [--description=<description>]
           Create a repository on github to push to"""
        root = self.gitm('rev-parse', '--show-toplevel').stdout.strip()
        name = os.path.basename(root)
        if opts['--org']:
            dest = self.gh.organization(opts['--org'])
            ns = opts['--org']
        else:
            dest = self.gh
            ns = self.my_login

        try:
            self.gh.repository(ns, name)
            err("Repository already exists")
        except github3.exceptions.NotFoundError:
            pass
        repo = dest.create_repository(name=name, description=opts['--description'] or "", private=opts['--private'])

        if 'origin' in self.remotes():
            print("Remote 'origin' already exists, adding the GitHub repository as 'github'")
            self.set_origin(opts, repo=repo, remote='github')
        else:
            self.set_origin(opts, repo=repo)

    @command
    def deploy_keys(self, opts):
        """[<repo>]
           Lists all keys for a repo"""
        repo = self.repository(opts)
        for key in repo.keys():
            ro = 'ro' if key.read_only else 'rw'
            print("%s %s (id: %s, %s)" % (key.key, key.title or '', key.id, ro))

    @command
    def edit_hook(self, opts):
        """<name> [<setting>...]
           Edit a hook"""
        for hook in self.repository(opts).hooks():
            if hook.name == opts['<name>']:
                break
        else:
            raise ValueError("Hook %s does not exist" % opts['<name>'])

        settings = dict([x.split('=', 1) for x in opts['<setting>']])
        for key in settings:
            if settings[key].isdigit():
                settings[key] = int(settings[key])
        events = settings.pop('events', ','.join(hook.events)).split(',')
        config = hook.config
        config.update(settings)
        hook.edit(config, events)

    @command
    def fetch(self, opts):
        """[--ssh|--http|--git] <user> [<refspec>]
           Fetch refs from a user's fork"""
        for fork in self.repository(opts).forks():
            if fork.owner.login in opts['<user>']:
                url = self.clone_url(fork, opts)
                refspec = opts['<refspec>'] or 'refs/heads/*'
                if ':' not in refspec:
                    if not refspec.startswith('refs/'):
                        refspec += ':' + 'refs/remotes/%s/' % fork.owner.login + refspec
                    else:
                        refspec += ':' + refspec.replace('refs/heads/', 'refs/remotes/%s/' % fork.owner.login)
                self.gitm('fetch', url, refspec, redirect=False)

    @command
    def fork(self, opts):
        """[--ssh|--http|--git] [--triangular [--upstream-branch=<branch>]] [<repo>]
           Fork a repo and clone it"""
        do_clone = bool(opts['<repo>'])
        repo = self.repository(opts)
        if repo.owner.login == self.my_login:
            err("You cannot fork your own repos")

        if isinstance(repo, github3.gists.Gist):
            for fork in repo.forks():
                if fork.owner.login == self.my_login:
                    err("You already forked this gist as %s" % fork.html_url)
        else:
            try:
                self.gh.repository(self.my_login, repo.name)
                err("Repository already exists")
            except github3.exceptions.NotFoundError:
                pass

        my_clone = repo.create_fork()
        self.wait_for_repo(my_clone.owner.login, my_clone.name, opts)
        if do_clone:
            self.clone(opts, repo=my_clone)
        else:
            self.set_origin(opts, repo=my_clone)

    @command
    @wants_parent
    def forks(self, opts):
        """[<repo>]
           List all forks of this repository"""
        repo = self.repository(opts)
        print("[%s] %s" % (wrap(repo.owner.login, attr.bright), repo.html_url))
        for fork in repo.forks():
            print("[%s] %s" % (fork.owner.login, fork.html_url))

    @command
    def gist(self, opts):
        """[--description=<description>] <file>...
           Create a new gist from files or stdin"""
        files = {}
        description = opts['--description'] or ''
        for f in opts['<file>']:
            if f == '-':
                files['stdout'] = {'content': sys.stdin.read()}
            else:
                if not os.path.exists(f):
                    err("No such file: %s" % f)
                with open(f) as fd:
                    files[os.path.basename(f)] = {'content': fd.read()}
        gist = self.gh.create_gist(description=description, files=files)
        print("Gist created at %s" % gist.html_url)

    @command
    def gists(self, opts):
        """[<user>]
           Show all gists for a user"""
        user = (opts['<user>'] or [self.gh.me().login])[0]
        for gist in self.gh.gists_by(user):
            print("%s - %s" % (gist.html_url, gist.description))

    @command
    def hooks(self, opts):
        """\nShow hooks that have been enabled"""
        for hook in self.repository(opts).hooks():
            print(wrap("%s (%s)" % (hook.name, ', '.join(hook.events)), attr.bright))
            for key, val in sorted(hook.config.items()):
                if val in (None, ''):
                    continue
                print("  %s: %s" % (key, val))

    @command
    def ignore(self, opts):
        """[<language>...]
           Show gitignore patterns for one or more languages"""
        lang = opts['<language>']
        if not lang:
            langs = sorted(self.gh.gitignore_templates(), key = lambda x: x.lower())
            print("Languages for which a gitignore template is available:\n  * " + "\n  * ".join(langs))
        else:
            langs = {}
            for l in lang:
                try:
                    t = self.gh.gitignore_template(langs.get(l.lower(), l))
                except github3.exceptions.NotFoundError:
                    if not langs:
                        langs = {x.lower(): x for x in self.gh.gitignore_templates()}
                    if l.lower() in langs:
                        t = self.gh.gitignore_template(langs[l.lower()])
                    else:
                        err("No gitignore template found for %s" % l)
                print("# Ignore patterns for " + l)
                print(t.strip())

    @command
    def ip_addresses(self, opts):
        """[--git] [--hooks] [--importer] [--pages]
           Show the IP addresses for github.com services in CIDR format"""
        ip_addresses = self.gh.meta()
        for what in ('git', 'hooks', 'importer', 'pages'):
            if opts['--' + what]:
                print("\n".join(ip_addresses[what]))

    @command
    def issue(self, opts):
        """[<repo>] [--parent] [<issue>...]
           Show issue details or report an issue"""
        if opts['<repo>'] and opts['<repo>'].isdigit():
            # Let's assume it's an issue
            opts['<issue>'].insert(0, opts['<repo>'])
            opts['<repo>'] = None
        repo = self.repository(opts)
        for issue_no in opts['<issue>']:
            issue = repo.issue(issue_no)
            if issue:
                pr = issue.pull_request()
                print(wrap(issue.title.encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding), attr.bright, attr.underline))
                print(issue.body.encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding))
                print(pr.html_url if pr else issue.html_url)
            else:
                print('No issue with id %s found in repository %s' % (issue_no, repo.full_name))
        if not opts['<issue>']:
            ext = ''
            template = self.find_template(repo, 'ISSUE_TEMPLATE')
            if template:
                if '.' in template[0]:
                    ext = template[0][template[0].rfind('.'):]
                body = template[1]
                extra = None
            else:
                body = ""
                extra = """Reporting an issue on %s/%s
Please describe the issue as clearly as possible. Lines starting with '#' will
be ignored, the first line will be used as title for the issue.""" % (repo.owner.login, repo.name)
            title, body = self.edit_msg(None, body, extra, 'ISSUE_EDITMSG' + ext)
            if not body:
                err("Empty issue message")

            try:
                issue = repo.create_issue(title=title, body=body)
                print("Issue %d created %s" % (issue.number, issue.html_url))
            except:
                filename = self.backup_message(title, body, 'issue-message-')
                err("Failed to create an issue, the issue text has been saved in %s" % filename)

    @command
    def issues(self, opts):
        """[<repo>] [--parent] [<filter>...]
           List issues in a repository"""
        if opts['<repo>'] and '=' in opts['<repo>']:
            opts['<filter>'].insert(0, opts['<repo>'])
            opts['<repo>'] = None
        if (not opts['<repo>'] and not self.in_repo) or opts['<repo>'] == '--':
            repos = list(self.gh.repositories(type='all'))
        else:
            repos = [self.repository(opts)]
        for repo in repos:
            repo = (opts['--parent'] and self.parent_repo(repo)) or repo
            filters = dict([x.split('=', 1) for x in opts['<filter>']])
            try:
                issues = list(repo.issues(**filters))
            except github3.GitHubError:
                _, err, _ = sys.exc_info()
                if err.code == 410:
                    if len(repos) == 1:
                        print(err.message)
                    continue
                else:
                    raise
            if not issues:
                continue
            print(wrap("Issues for %s/%s" % (repo.owner.login, repo.name), attr.bright))
            for issue in issues:
                pr = issue.pull_request()
                url = pr.html_url if pr else issue.html_url
                print("[%d] %s %s" % (issue.number, issue.title.encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding), url))

    @command
    def log(self, opts):
        """[--type=<type>] [--count=<count>] [--verbose] [<what>]
           Display github log for yourself or other users. Or for an organisation or a repo"""
        logtype = 'user'
        count = int(opts['--count'] or 30)
        verbose = opts['--verbose']
        if not opts['<what>']:
            what = self.me
        else:
            if '/' in opts['<what>']:
                logtype = 'repo'
                user, repo = opts['<what>'].split('/', 1)
                if user == 'gist':
                    what = self.gh.gist(repo)
                    if not what:
                        err("Gist %s does not exist" % repo)
                else:
                    what = self.gh.repository(user, repo)
                    if not what:
                        err("Repository %s/%s does not exist" % (user, repo))
            else:
                what = self.gh.user(opts['<what>'])
                if what.type == 'Organization':
                    logtype = 'org'
                if not what:
                    err("User %s does not exist" % opts['<what>'])

        if not opts['--type']:
            events = [x for x in what.events(number=count)]
        else:
            events = []
            etype = opts['--type'].lower() + 'event'
            for event in what.events(number=-1):
                if event.type.lower() == etype:
                    events.append(event)
                    if len(events) == count:
                        break

        now = datetime.datetime.now()
        for event in reversed(events):
            ts = event.created_at
            if ts.year == now.year:
                if (ts.month, ts.day) == (now.month, now.day):
                    ts = wrap(ts.strftime("%H:%M"), attr.faint)
                    tss = '     '
                else:
                    ts = wrap(ts.strftime("%m/%d %H:%M"), attr.faint)
                    tss = '           '
            else:
                ts = wrap(ts.strftime("%Y/%m/%d %H:%M"), attr.faint)
                tss = '                '
            repo = event.repo['name']
            repo_ = ' (%s)' % repo
            if logtype != 'user':
                repo_ = ''
                ts += ' %s' % event.actor.login
            if event.type == 'CommitCommentEvent':
                print("%s commented on commit %s%s" % (ts, event.payload['comment'].commit_id[:7], repo_))
            elif event.type == 'CreateEvent':
                if event.payload['ref_type'] == 'repository':
                    print("%s created %s %s" % (ts, event.payload['ref_type'], repo))
                else:
                    print("%s created %s %s%s" % (ts, event.payload['ref_type'], event.payload['ref'], repo_))
            elif event.type == 'DeleteEvent':
                print("%s deleted %s %s%s" % (ts, event.payload['ref_type'], event.payload['ref'], repo_))
            elif event.type == 'DownloadEvent':
                print("%s created download %s (%s)" % (ts, event.payload['name'], event.payload['description']))
            elif event.type == 'FollowEvent':
                print("%s started following %s" % (ts, event.payload['target'].login))
            elif event.type == 'ForkEvent':
                print("%s forked %s to %s/%s" % (ts, repo, event.payload['forkee'].owner.login, event.payload['forkee'].name))
            elif event.type == 'ForkApplyEvent':
                print("%s applied %s to %s%s" % (ts, event.payload['after'][:7], event.payload['head'], repo_))
            elif event.type == 'GistEvent':
                print("%s %sd gist #%s" % (ts, event.payload['action'], event.payload['gist'].html_url))
            elif event.type == 'GollumEvent':
                pages = len(event.payload['pages'])
                print("%s updated %d wikipage%s%s" % (ts, pages, {1:''}.get(pages, 's'), repo_))
            elif event.type == 'IssueCommentEvent':
                print("%s commented on issue #%s%s" % (ts, event.payload['issue'].number, repo_))
                if verbose:
                    print("%s %s %s" % (tss, event.payload['issue'].title, event.payload['comment'].html_url))
            elif event.type == 'IssuesEvent':
                print("%s %s issue #%s%s" % (ts, event.payload['action'], event.payload['issue'].number, repo_))
                if verbose:
                    print("%s %s %s" % (tss, event.payload['issue'].title, event.payload['issue'].html_url))
            elif event.type == 'MemberEvent':
                print("%s %s %s to %s" % (ts, event.payload['action'], event.payload['member'].login, repo))
            elif event.type == 'PublicEvent':
                print("%s made %s open source" % repo)
            elif event.type == 'PullRequestReviewCommentEvent':
                print("%s commented on a pull request for commit %s%s" % (ts, event.payload['comment'].commit_id[:7], repo_))
            elif event.type == 'PullRequestEvent':
                print("%s %s pull_request #%s%s" % (ts, event.payload['action'], event.payload['pull_request'].number, repo_))
                if verbose:
                    print("%s %s %s" % (tss, event.payload['pull_request'].title, event.payload['pull_request'].html_url))

            elif event.type == 'PushEvent':
                # Old push events have shas and not commits
                if 'commits' in event.payload:
                    commits = len(event.payload['commits'])
                else:
                    commits = len(event.payload['shas'])
                print("%s pushed %d commits to %s%s" % (ts, commits, event.payload['ref'][11:], repo_))
                if verbose:
                    shas = '%s...%s' % (event.payload['before'][:8], event.payload['head'][:8])
                    print("%s %s/%s/compare/%s" % (tss, self.me.html_url, event.repo[1], shas))
            elif event.type == 'ReleaseEvent':
                print("%s released %s" % (ts, event.payload['name']))
            elif event.type == 'StatusEvent':
                print("%s commit %s changed to %s" % (ts, event.payload['sha'][:7], event.payload['state']))
            elif event.type == 'TeamAddEvent':
                if 'user' in event.payload:
                    what = 'user'
                    name = isinstance(event.payload['user'], dict) and event.payload['user']['name'] or event.payload['user'].name
                else:
                    what = 'repository'
                    name = isinstance(event.payload['repository'], dict) and event.payload['repository']['name'] or event.payload['repository'].name
                    print("%s %s %s was added to team %s" % (ts, what, name, event.payload['team'].name))
            elif event.type == 'WatchEvent':
                print("%s %s watching %s" % (ts, event.payload['action'], repo))
            elif event.type == 'GistHistoryEvent':
                print("%s committed %s additions, %s deletions" % (ts, event.additions, event.deletions))
            else:
                print(wrap("Cannot display %s. Please file a bug at github.com/seveas/git-spindle\nincluding the following output:" % event.type, attr.bright))
                pprint(event.payload)

    @command
    def ls(self, opts):
        """[<dir>...]
           Display the contents of a directory on GitHub"""
        for arg in opts['<dir>'] or ['']:
            repo, ref, file = ([None, None] + arg.split(':',2))[-3:]
            user = None
            if repo:
                user, repo = ([None] + repo.split('/'))[-2:]
                repo = self.gh.repository(user or self.my_login, repo)
            else:
                repo = self.repository(opts)
                file = self.rel2root(file)
            content = repo.directory_contents(file, ref=ref or repo.default_branch)
            if not content:
                err("No such directory: %s" % arg)
            content = sorted([v for k, v in content], key=lambda file: file.name)
            mt = max([len(file.type) for file in content])
            ms = max([len(str(file.size)) for file in content])
            fmt = "%%(type)-%ds %%(size)-%ds %%(sha).7s %%(path)s" % (mt, ms)
            for file in content:
                print(fmt % file.__dict__)

    @command
    def mirror(self, opts):
        """[--ssh|--http|--git] [<repo>]
           Mirror a repository, or all repositories for a user"""
        if opts['<repo>'] and opts['<repo>'].endswith('/*'):
            user = opts['<repo>'].rsplit('/', 2)[-2]
            for repo in self.gh.repositories(type='all') if user == self.my_login else self.gh.repositories_by(user, type='all'):
                if repo.owner.login != self.my_login:
                    continue
                opts['<repo>'] = '%s/%s' % (user, repo)
                self.mirror(opts)
            for repo in self.gh.gists_by(user):
                opts['<repo>'] = 'gist/%s' % repo.name
                self.mirror(opts)
            return
        repo = self.repository(opts)
        git_dir = repo.name + '.git'
        cur_dir = os.path.basename(os.path.abspath(os.getcwd()))
        if cur_dir != git_dir and not os.path.exists(git_dir):
            url = self.clone_url(repo, opts)
            self.gitm('clone', '--mirror', url, redirect=False)
        else:
            if git_dir == cur_dir:
                git_dir = '.'
            # Update the current, mirrored repo
            if self.git('--git-dir', git_dir, 'config', 'core.bare').stdout.strip() != 'true' or \
               self.git('--git-dir', git_dir, 'config', 'remote.origin.mirror').stdout.strip() != 'true':
                   err("This is not a mirrored repository")
            self.gitm('--git-dir', git_dir, 'fetch', '-q', '--prune', 'origin', redirect=False)

        with open(os.path.join(git_dir, 'description'), 'w') as fd:
            fd.write(repo.description or "")

    @command
    def network(self, opts):
        """[<level>]
           Create a graphviz graph of followers and forks"""
        from collections import defaultdict
        class P:
            def __init__(self, user):
                self.user = user
                self.done = False
                self.rel_to = defaultdict(list)

            def __repr__(self):
                return dict.__repr__(self.rel_to)

        level = 1
        if opts['<level>']:
            try:
                level = int(opts['<level>'])
            except ValueError:
                err("Integer argument required")
        people = {self.my_login: P(self.me)}
        for i in range(level):
            for login, person in list(people.items()):
                if person.done:
                    continue

                sys.stderr.write("Looking at user %s\n" % login)
                # Followers
                for other in person.user.followers():
                    if other.login not in people:
                        people[other.login] = P(other)
                    people[other.login].rel_to[login].append('follows')
                for other in person.user.following():
                    if other.login not in people:
                        people[other.login] = P(other)
                    person.rel_to[other.login].append('follows')

                # Forks
                for repo in self.gh.repositories_by(login, type='owner'):
                    sys.stderr.write("Looking at repo %s\n" % repo.name)
                    if repo.fork:
                        if repo.owner.login not in people:
                            people[repo.owner.login] = P(repo.owner)
                        parent = self.parent_repo(repo)
                        person.rel_to[parent.owner.login].append('forked %s' % parent.name)
                    else:
                        for fork in repo.forks():
                            if fork.owner.login == login:
                                continue
                            if fork.owner.login not in people:
                                people[fork.owner.login] = P(fork.owner)
                            people[fork.owner.login].rel_to[login].append('forked %s' % repo.name)
                person.done = True

        # Now we create a graph
        graph = ["digraph network {"]
        for person in people:
            graph.append('    "%s"' % person)

        for login, person in people.items():
            for other, types in person.rel_to.items():
                graph.append('    "%s" -> "%s" [label="%s"]' % (login, other, "\\n".join(types)))

        graph.append("}")
        print("\n".join(graph))

    @command
    def protect(self, opts):
        """[--enforcement=<level>] [--status-checks=<contexts>] <branch>
           Protect a branch against deletions, force-pushes and failed status checks"""
        repo = self.repository(opts)
        repo.branch(opts['<branch>']).protect(enforcement=opts['--enforcement'],
                                              status_checks=(opts['--status-checks'] or '').split(','))

    @command
    def protected(self, opts):
        """\nList active branch protections"""
        repo = self.repository(opts)
        for branch in repo.branches(protected=True):
            data = branch.protection
            msg = branch.name
            if data['required_status_checks']['contexts'] and data['required_status_checks']['enforcement_level'] != 'off':
                msg += ' (%s must pass for %s)' % (','.join(data['required_status_checks']['contexts']), data['required_status_checks']['enforcement_level'])
            print(msg)

    @command
    def public_keys(self, opts):
        """[<user>]
           Lists all keys for a user"""
        user = opts['<user>'] and opts['<user>'][0] or self.my_login
        if self.my_login == user:
            keys = self.gh.keys()
        else:
            keys = self.gh.user(user).keys()
        for key in keys:
            print("%s %s" % (key.key, getattr(key, 'title', '')))


    @command
    def pull_request(self, opts):
        """[--issue=<issue>] [--yes] [<yours:theirs>]
           Opens a pull request to merge your branch to an upstream branch"""
        repo = self.repository(opts)
        parent = self.parent_repo(repo) or repo
        # Which branch?
        src = opts['<yours:theirs>'] or ''
        dst = None
        if ':' in src:
            src, dst = src.split(':', 1)
        if not src:
            src = self.gitm('rev-parse', '--abbrev-ref', 'HEAD').stdout.strip()
        if not dst:
            dst = parent.default_branch
            tracking_branch = self.git('rev-parse', '--symbolic-full-name', '%s@{u}' % src).stdout.strip()
            if tracking_branch.startswith('refs/remotes/'):
                tracking_remote, tracking_branch = tracking_branch.split('/', 3)[-2:]
                if tracking_branch != src or repo.remote != tracking_remote:
                    # Interesting. We're not just tracking a branch in our clone!
                    dst = tracking_branch

        if src == dst and parent == repo:
            err("Cannot file a pull request on the same branch")

        # Try to get the local commit
        commit = self.gitm('show-ref', 'refs/heads/%s' % src).stdout.split()[0]
        # Do they exist on github?
        try:
            srcb = repo.branch(src)
        except github3.exceptions.NotFoundError:
            srcb = None
        if not srcb:
            if self.question("Branch %s does not exist in your GitHub repo, shall I push?" % src):
                self.gitm('push', '-u', repo.remote, src, redirect=False)
            else:
                err("Aborting")
        elif srcb and srcb.commit.sha != commit:
            # Have we diverged? Then there are commits that are reachable from the github branch but not local
            diverged = self.gitm('rev-list', srcb.commit.sha, '^' + commit)
            if diverged.stderr or diverged.stdout:
                if self.question("Branch %s has diverged from GitHub, shall I push and overwrite?" % src, default=False):
                    self.gitm('push', '--force', repo.remote, src, redirect=False)
                else:
                    err("Aborting")
            else:
                if self.question("Branch %s not up to date on github, but can be fast forwarded, shall I push?" % src):
                    self.gitm('push', repo.remote, src, redirect=False)
                else:
                    err("Aborting")

        dstb = parent.branch(dst)
        if not dstb:
            err("Branch %s does not exist in %s/%s" % (dst, parent.owner.login, parent.name))

        # Do we have the dst locally?
        for remote in self.gitm('remote').stdout.strip().split("\n"):
            url = self.gitm('config', 'remote.%s.url' % remote).stdout.strip()
            if url in [parent.git_url, parent.ssh_url, parent.clone_url]:
                if parent.private and url != parent.ssh_url:
                    err("You should configure %s/%s to fetch via ssh, it is a private repo" % (parent.owner.login, parent.name))
                self.gitm('fetch', remote, redirect=False)
                break
        else:
            err("You don't have %s/%s configured as a remote repository" % (parent.owner.login, parent.name))

        # How many commits?
        accept_empty_body = False
        commits = self.gitm('log', '--pretty=%H', '%s/%s..%s' % (remote, dst, src)).stdout.strip().split()
        commits.reverse()
        if not commits:
            err("Your branch has no commits yet")
        # Are we turning an issue into a commit?
        if opts['--issue']:
            pull = parent.create_pull_from_issue(base=dst, head='%s:%s' % (repo.owner.login, src), issue=int(opts['--issue']))
            print("Pull request %d created %s" % (pull.number, pull.html_url))
            return

        # 1 commit: title/body from commit
        if len(commits) == 1:
            title, body = self.gitm('log', '--pretty=%s\n%b', '%s^..%s' % (commits[0], commits[0])).stdout.split('\n', 1)
            title = title.strip()
            body = body.strip()
            accept_empty_body = not bool(body)

        # More commits: title from branchname (titlecased, s/-/ /g), body comments from shortlog
        else:
            title = src
            if '/' in title:
                title = title[title.rfind('/') + 1:]
            title = title.title().replace('-', ' ')
            body = ""

        ext = ''
        template = self.find_template(repo, 'PULL_REQUEST_TEMPLATE')
        if template:
            if '.' in template[0]:
                ext = template[0][template[0].rfind('.'):]
            body = template[1].rstrip() + '\n\n' + body

        extra = """Requesting a pull from %s/%s into %s/%s

Please enter a message to accompany your pull request. Lines starting
with '#' will be ignored, and an empty message aborts the request.""" % (repo.owner.login, src, parent.owner.login, dst)
        extra += "\n\n " + self.gitm('shortlog', '%s/%s..%s' % (remote, dst, src)).stdout.strip()
        extra += "\n\n " + self.gitm('diff', '--stat', '%s^..%s' % (commits[0], commits[-1])).stdout.strip()
        title, body = self.edit_msg(title, body, extra, 'PULL_REQUEST_EDIT_MSG' + ext)
        if not body and not accept_empty_body:
            err("No pull request message specified")

        try:
            pull = parent.create_pull(base=dst, head='%s:%s' % (repo.owner.login, src), title=title, body=body)
            print("Pull request %d created %s" % (pull.number, pull.html_url))
        except:
            filename = self.backup_message(title, body, 'pull-request-message-')
            err("Failed to create a pull request, the pull request text has been saved in %s" % filename)

    @command
    def readme(self, opts):
        """[<repo>]
           Get the README for a repository"""
        repo = self.repository(opts)
        readme = repo.readme()
        if readme:
            os.write(sys.stdout.fileno(), readme.decoded)
        else:
            err("No readme found")

    @command
    def release(self, opts):
        """[--draft] [--prerelease] <tag> [<releasename>]
           Create a release"""
        repo = self.repository(opts)
        tag = opts['<tag>']
        if tag.startswith('refs/tags/'):
            tag = tag[10:]
        name = opts['<releasename>'] or tag
        ref = 'refs/tags/' + tag
        ret = self.git('rev-parse', '--quiet', '--verify', ref + '^0')
        if not ret:
            err("Tag %s does not exist yet" % tag)
        sha = ret.stdout.strip()
        if not self.git('ls-remote', repo.remote, ref).stdout.strip():
            if self.question("Tag %s does not exist in your GitHub repo, shall I push?" % tag):
                self.gitm('push', repo.remote, '%s:%s' % (ref, ref), redirect=False)
        body = ''
        if self.git('cat-file', '-t', ref).stdout.strip() == 'tag':
            body = self.git('--no-pager', 'log', '-1', '--format=%B', ref).stdout
        extra = """Creating release %s based on tag %s
Please enter a text to accompany your release. Lines starting with '#'
will be ignored""" % (name, tag)
        body = self.edit_msg(None, body, extra, 'RELEASE_TEXT', split_title=False)
        release = repo.create_release(tag, target_commitish=sha, name=name, body=body, draft=opts['--draft'], prerelease=opts['--prerelease'])
        print("Release '%s' created %s" % (release.name, release.html_url))

    @command
    def releases(self, opts):
        """[<repo>]
           List all releases"""
        repo = self.repository(opts)
        for release in repo.releases():
            status = []
            if release.draft:
                status.append('draft')
            if release.prerelease:
                status.append('prerelease')
            status = status and ' ' + wrap(','.join(status), attr.faint) or ''
            print("%s (%s)%s %s" % (release.name, release.tag_name, status, release.html_url))

    @command
    def remove_collaborator(self, opts):
        """<user>...
           Remove a user as collaborator """
        repo = self.repository(opts)
        for user in opts['<user>']:
            repo.remove_collaborator(user)

    @command
    def remove_deploy_key(self, opts):
        """<key>...
           Remove deploy key by id"""
        repo = self.repository(opts)
        for key in opts['<key>']:
            if not key.isdigit():
                err("You must specify a numeric id")
            try:
                repo.key(key).delete()
            except github3.exceptions.NotFoundError:
                err("No key with id %s found" % key)

    @command
    def remove_hook(self, opts):
        """<name>
           Remove a hook"""
        for hook in self.repository(opts).hooks():
            if hook.name == opts['<name>']:
                hook.delete()

    @command
    def render(self, opts):
        """[--save=<outfile>] <file>
           Render a markdown document"""
        template = """<!DOCTYPE html>
<html>
  <head>
    <link type="text/css" rel="stylesheet" media="all" href="http://necolas.github.io/normalize.css/latest/normalize.css"></link>
    <link type="text/css" rel="stylesheet" media="all" href="http://seveas.github.io/git-spindle/_static/github.css"></link>
    <link type="text/css" rel="stylesheet" media="all" href="https://cdnjs.cloudflare.com/ajax/libs/octicons/2.0.2/octicons.css"></link>
  </head>
  <body>
    <div class="container">
      <div id="readme" class="boxed-group">
        <h3><span class="octicon octicon-book"></span> %s</h3>
        <article class="markdown-body">
          %s
        </article>
      </div>
    </div>
  </body>
</html>"""
        with open(opts['<file>'][0]) as fd:
            data = fd.read()
        rendered = github3.markdown(data)
        if isinstance(rendered, bytes):
            rendered = rendered.decode('utf-8')
        rendered = rendered.replace('user-content-', '')
        html = template % (os.path.basename(opts['<file>'][0]), rendered)
        if opts['--save']:
            with open(opts['--save'], 'w') as fd:
                fd.write(html)
        else:
            with tempfile.NamedTemporaryFile(suffix='.html', delete=False) as fd:
                fd.write(html.encode('utf-8'))
                fd.close()
                webbrowser.open('file://' + fd.name)
                time.sleep(1)
                os.remove(fd.name)

    @command
    def repos(self, opts):
        """[--no-forks] [<user>]
           List all repos of a user, by default yours"""
        if opts['<user>']:
            repos = list(self.gh.repositories_by(opts['<user>'][0]))
        else:
            repos = list(self.gh.repositories(type='all'))
            opts['<user>'] = [self.my_login]
        if not repos:
            return
        maxlen = max([len(x.name) for x in repos])
        maxstar = len(str(max([x.stargazers_count for x in repos])))
        maxfork = len(str(max([x.forks_count for x in repos])))
        maxwatch = len(str(max([x.watchers for x in repos])))
        # XXX github support request filed: watchers is actually stars
        #fmt = u"%%-%ds \u2605 %%-%ds \u25c9 %%-%ds \u2919 %%-%ds %%s" % (maxlen, maxstar, maxwatch, maxfork)
        fmt = u"%%-%ds \u2605 %%-%ds \u2919 %%-%ds %%s" % (maxlen, maxstar, maxfork)
        for repo in repos:
            color = [attr.normal]
            if repo.private:
                color.append(fgcolor.red)
            if repo.fork:
                if opts['--no-forks']:
                    continue
                color.append(attr.faint)
            name = repo.name
            if opts['<user>'][0] != repo.owner.login:
                name = '%s/%s' % (repo.owner.login, name)
            msg = wrap(fmt % (name, repo.stargazers_count, repo.forks_count, repo.description), *color)
            print(msg)

    @command
    def say(self, opts):
        """[<msg>]
           Let the octocat speak to you"""
        msg = github3.octocat(opts['<msg>'] or None)
        if isinstance(msg, bytes):
            msg = msg.decode('utf-8')
        print(msg)

    @command
    def set_origin(self, opts, repo=None, remote='origin'):
        """[--ssh|--http|--git] [--triangular [--upstream-branch=<branch>]]
           Set the remote 'origin' to github.
           If this is a fork, set the remote 'upstream' to the parent"""
        if not repo:
            repo = self.repository(opts)
            # Is this mine? No? Do I have a clone?
            if repo.owner.login != self.my_login:
                my_repo = self.gh.repository(self.me, repo.name)
                if my_repo:
                    repo = my_repo

        url = self.clone_url(repo, opts)
        if self.git('config', 'remote.%s.url' % remote).stdout.strip() != url:
            print("Pointing %s to %s" % (remote, url))
            self.gitm('config', 'remote.%s.url' % remote, url)
        self.gitm('config', '--replace-all', 'remote.%s.fetch' % remote, '+refs/heads/*:refs/remotes/%s/*' % remote)

        if repo.fork:
            parent = self.parent_repo(repo)
            url = self.clone_url(parent, opts)
            if self.git('config', 'remote.upstream.url').stdout.strip() != url:
                print("Pointing upstream to %s" % url)
                self.gitm('config', 'remote.upstream.url', url)
            self.gitm('config', 'remote.upstream.fetch', '+refs/heads/*:refs/remotes/upstream/*')
        else:
            # If issues are enabled, fetch pull requests
            try:
                list(repo.issues(number=1))
            except github3.GitHubError:
                pass
            else:
                self.gitm('config', '--add', 'remote.%s.fetch' % remote, '+refs/pull/*/head:refs/pull/*/head')

        if self.git('ls-remote', remote).stdout.strip():
            self.gitm('fetch', remote, redirect=False)
        if repo.fork:
            self.gitm('fetch', 'upstream', redirect=False)

        if remote != 'origin':
            return

        self.set_tracking_branches(remote, upstream="upstream", triangular=opts['--triangular'], upstream_branch=opts['--upstream-branch'])

    @command
    def status(self, opts):
        """\nDisplay current and historical GitHub service status"""
        api = github3.GitHubStatus()
        messages = api.messages()
        if not messages:
            messages = [api.last_message()]
        status = api.status()
        status.update({
            'body': 'Current status: %s' % status['status'],
            'created_on': status['last_updated'],
        })
        messages.insert(0, status)
        for message in reversed(messages):
            ts = time.strptime(message['created_on'], '%Y-%m-%dT%H:%M:%SZ')
            offset = time.timezone
            if time.daylight:
                offset = time.altzone
            color = {'good': fgcolor.green, 'minor': fgcolor.yellow, 'major': fgcolor.red}[message['status']]
            ts = datetime.datetime(ts.tm_year, ts.tm_mon, ts.tm_mday, ts.tm_hour, ts.tm_min, ts.tm_sec) - datetime.timedelta(0,offset)
            print('%s %s %s' % (wrap(ts.strftime('%Y-%m-%d %H:%M'), attr.faint), wrap("%-5s" % message['status'], color), message['body']))

    @command
    def unprotect(self, opts):
        """ <branch>
           Remove branch protections from a branch"""
        repo = self.repository(opts)
        repo.branch(opts['<branch>']).unprotect()

    @command
    def whoami(self, opts):
        """\nDisplay GitHub user info"""
        opts['<user>'] = [self.my_login]
        self.whois(opts)

    @command
    def whois(self, opts):
        """<user>...
           Display GitHub user info"""
        for user_ in opts['<user>']:
            user = self.gh.user(user_)
            if not user:
                print("No such user: %s" % user_)
                continue
            emails = {}
            if user.login == self.my_login:
                for email in self.gh.emails():
                    emails[email.email] = email
            print(wrap(user.name or user.login, attr.bright, attr.underline))
            print('Profile   %s' % user.html_url)
            if user.email:
                unverified = ''
                if user.email in emails and not emails[user.email].verified:
                    unverified = ' ' + wrap('(not verified)', fgcolor.red, attr.bright)
                print('Email     %s%s' % (user.email, unverified))
                for email in emails:
                    if email == user.email:
                        continue
                    unverified = ''
                    if not emails[email].verified:
                        unverified = ' ' + wrap('(not verified)', fgcolor.red, attr.bright)
                    print('          %s%s' % (email, unverified))
            if user.blog:
                print('Blog      %s' % user.blog)
            if user.location:
                print('Location  %s' % user.location)
            if user.company:
                print('Company   %s' % user.company)
            print('Repos     %d' % user.public_repos_count)
            print('Gists     %d' % user.public_gists_count)
            if user.login == self.my_login:
                keys = self.gh.keys()
            else:
                keys = user.keys()
            for pkey in keys:
                algo, key = pkey.key.split()[:2]
                algo = algo[4:].upper()
                if getattr(pkey, 'title', None):
                    print("%s key%s...%s (%s)" % (algo, ' ' * (6 - len(algo)), key[-10:], pkey.title))
                else:
                    print("%s key%s...%s" % (algo, ' ' * (6 - len(algo)), key[-10:]))
            if user.login == self.my_login:
                orgs = self.gh.organizations()
            else:
                orgs = list(user.organizations())
            if orgs:
                print("Member of %s" % ', '.join([x.login for x in orgs]))
            if user.type == 'Organization':
                print('Members:')
                for member in self.gh.organization(user.login).members():
                    print(" - %s" % member.login)

    def graphql_method(self, name, query):
        def method(**args):
            url = self.gh._build_url('graphql')
            args = {'query': query, 'variables': args}
            resp = self.gh._post(url, data=args)
            data = self.gh._json(resp, 200)
            if data.get('errors', []):
                raise GraphqlError(data['errors'])
            return graphql_fixup(data['data'])
        setattr(self, name, method)

class GraphqlError(Exception):
    pass

class GraphqlObject:
    def __init__(self, attrs):
        for k, v in attrs.items():
            setattr(self, k, graphql_fixup(v))

def graphql_fixup(data):
    if isinstance(data, dict):
        if 'nodes' in data and len(data.keys()) == 1:
            return graphql_fixup(data['nodes'])
        return GraphqlObject(data)
    if isinstance(data, list):
        data = [graphql_fixup(x) for x in data]
    return data



from gitspindle import *
from gitspindle.ansi import *
import gitspindle.bbapi as bbapi
import getpass
import glob
import os
import sys
import webbrowser
import binascii

class BitBucket(GitSpindle):
    prog = 'git bucket'
    what = 'BitBucket'
    spindle = 'bitbucket'
    hosts = ['bitbucket.org', 'www.bitbucket.org']
    api = bbapi

    # Support functions
    def login(self):
        user = self.config('user')
        if not user:
            user = input("BitBucket user: ").strip()
            self.config('user', user)

        password = self.config('password')
        if not password:
            print("Generate an application password at https://bitbucket.org/account/settings/app-passwords/ and enter it below")
            password = getpass.getpass("BitBucket password: ")
            try:
                bbapi.Bitbucket(user, password).user(user)
            except:
                err("Authentication failed")
            self.config('password', password)
            location = '%s - do not share this file' % self.config_file
            if self.use_credential_helper:
                location = 'git\'s credential helper'
            print("Your BitBucket authentication password is now stored in %s" % location)

        self.bb = bbapi.Bitbucket(user, password)
        self.me = self.bb.user(user)
        self.my_login = self.me.nickname

    def parse_url(self, url):
        return ([self.my_login] + url.path.split('/'))[-2:]

    def get_repo(self, remote, workspace, repo):
        try:
            return self.bb.workspace(workspace).repository(repo)
        except bbapi.BitBucketError:
            pass

    def parent_repo(self, repo):
        if repo.parent:
            return bbapi.Repository(self.bb, url=repo.parent['links']['self']['href'])

    def clone_url(self, repo, opts):
        if opts['--ssh'] or repo.is_private:
            return repo.links['clone']['ssh']
        if opts['--http']:
            return repo.links['clone']['https']
        if repo.owner['nickname'] == self.my_login:
            return repo.links['clone']['ssh']
        return repo.links['clone']['https']

    def api_root(self):
        return 'https://bitbucket.org/api/'

    # Commands
    @command
    def add_deploy_key(self, opts):
        """<key>...
           Add a deploy key"""
        repo = self.repository(opts)
        for arg in opts['<key>']:
            with open(arg) as fd:
                algo, key, label = fd.read().strip().split(None, 2)
            key = "%s %s" % (algo, key)
            print("Adding deploy key %s" % arg)
            repo.add_deploy_key(key, label)

    @command
    def add_public_keys(self, opts):
        """[<key>...]
           Adds keys to your public keys"""
        if not opts['<key>']:
            opts['<key>'] = glob.glob(os.path.join(os.path.expanduser('~'), '.ssh', 'id_*.pub'))
        existing = [x.key for x in self.me.keys()]
        for arg in opts['<key>']:
            with open(arg) as fd:
                algo, key, title = fd.read().strip().split(None, 2)
            key = "%s %s" % (algo, key)
            if key in existing:
                continue
            print("Adding %s" % arg)
            self.me.create_key(label=title, key=key)

    @command
    @wants_parent
    def add_remote(self, opts):
        """[--ssh|--http] <user> [<name>]
           Add user's fork as a named remote. The name defaults to the user's loginname"""
        for fork in self.repository(opts).forks():
            if fork.owner['nickname'] in opts['<user>']:
                name = opts['<name>'] or fork.owner['nickname']
                url = self.clone_url(fork, opts)
                self.gitm('remote', 'add', '-f', name, url, redirect=False)

    @command
    @wants_parent
    def apply_pr(self, opts):
        """[--ssh|--http] <pr-number>
           Applies a pull request as a series of cherry-picks"""
        repo = self.repository(opts)
        pr = repo.pull_request(opts['<pr-number>'])
        if not pr:
            err("Pull request %s does not exist" % opts['<pr-number>'])
        print("Applying PR#%d from %s: %s" % (pr.id, pr.author['display_name'] or pr.author['nickname'], pr.title))
        # Warnings
        warned = False
        cbr = self.gitm('rev-parse', '--symbolic-full-name', 'HEAD').stdout.strip().replace('refs/heads/','')
        if cbr != pr.destination['branch']['name']:
            print(wrap("Pull request was filed against %s, but you're on the %s branch" % (pr.destination['branch']['name'], cbr), fgcolor.red))
            warned = True
        if pr.state == 'MERGED':
            print(wrap("Pull request was already merged by %s" % (pr.closed_by['display_name'] or pr.closed_by['nickname']), fgcolor.red))
            warned = True
        if pr.state == 'DECLINED':
            print(wrap("Pull request has already been declined: %s" % pr.reason, fgcolor.red))
            warned = True
        if warned:
            if not self.question("Continue?"):
                sys.exit(1)

        # Fetch PR if needed
        sha = self.git('rev-parse', '--verify', 'refs/pull/%d/head' % pr.id).stdout.strip()
        if not sha.startswith(pr.source['commit']['hash']):
            print("Fetching pull request")
            ws, repo = pr.source['repository']['full_name'].split('/')
            url = self.clone_url(self.bb.workspace(ws).repository(repo), opts)
            self.gitm('fetch', url, 'refs/heads/%s:refs/pull/%d/head' % (pr.source['branch']['name'], pr.id), redirect=False)
        head_sha = self.gitm('rev-parse', 'HEAD').stdout.strip()
        if self.git('merge-base', pr.source['commit']['hash'], head_sha).stdout.strip() == head_sha:
            print("Fast-forward merging %s..refs/pull/%d/head" % (pr.destination['branch']['name'], pr.id))
            self.gitm('merge', '--ff-only', 'refs/pull/%d/head' % pr.id, redirect=False)
        else:
            print("Cherry-picking  %s..refs/pull/%d/head" % (pr.destination['branch']['name'], pr.id))
            self.gitm('cherry-pick', '%s..refs/pull/%d/head' % (pr.destination['branch']['name'], pr.id), redirect=False)

    @command
    def browse(self, opts):
        """[--parent] [<repo>] [<section>]
           Open the Bitbucket page for a repository in a browser"""
        sections = ['src', 'commits', 'branches', 'pull-requests', 'downloads', 'admin', 'issues', 'wiki']
        if opts['<repo>'] in sections and not opts['<section>']:
            opts['<repo>'], opts['<section>'] = None, opts['<repo>']
        repo = self.repository(opts)
        url = repo.links['html']['href']
        if opts['<section>']:
            url += '/' + opts['<section>']
        webbrowser.open_new_tab(url)

    @command
    def cat(self, opts):
        """<file>...
           Display the contents of a file on BitBucket"""
        for arg in opts['<file>']:
            repo, ref, file = ([None, None] + arg.split(':',2))[-3:]
            user = None
            if repo:
                user, repo = ([None] + repo.split('/'))[-2:]
                repo = self.bb.repository(user or self.my_login, repo)
            else:
                repo = self.repository(opts)
                file = self.rel2root(file)
            try:
                content = repo.src(path=file, revision=ref or 'master') # BitBucket has no API to retrieve the default branch
            except bbapi.BitBucketError:
                err("No such file: %s" % arg)
            if not isinstance(content, bytes):
                err("Not a regular file: %s" % arg)
            os.write(sys.stdout.fileno(), content)

    @command
    def clone(self, opts, repo=None):
        """[--ssh|--http] [--triangular [--upstream-branch=<branch>]] [--parent] [git-clone-options] <repo> [<dir>]
           Clone a repository by name"""
        if not repo:
            repo = self.repository(opts)
        url = self.clone_url(repo, opts)

        args = opts['extra-opts']
        args.append(url)
        dir = opts['<dir>'] and opts['<dir>'][0] or repo.name
        if '--bare' in args:
            dir += '.git'
        args.append(dir)

        self.gitm('clone', *args, redirect=False).returncode
        if repo.parent:
            os.chdir(dir)
            self.set_origin(opts, repo=repo)
            self.gitm('fetch', 'upstream', redirect=False)

    @command
    def create(self, opts):
        """[--private] [--team=<team>] [--description=<description>]
           Create a repository on bitbucket to push to"""
        root = self.gitm('rev-parse', '--show-toplevel').stdout.strip()
        name = os.path.basename(root)
        if opts['--team']:
            dest = self.bb.team(opts['--team'])
        else:
            dest = self.me
        try:
            dest.repository(name)
            err("Repository already exists")
        except bbapi.BitBucketError:
            pass

        repo = dest.create_repository(slug=name, description=opts['--description'], is_private=opts['--private'], has_issues=True, has_wiki=True)
        if 'origin' in self.remotes():
            print("Remote 'origin' already exists, adding the BitBucket repository as 'bitbucket'")
            self.set_origin(opts, repo=repo, remote='bitbucket')
        else:
            self.set_origin(opts, repo=repo)

    @command
    def deploy_keys(self, opts):
        """[<repo>]
           Lists all keys for a repo"""
        repo = self.repository(opts)
        for key in repo.deploy_keys():
            print("%s %s (id: %s)" % (key['key'], key.get('label', ''), key['id']))

    @command
    def fetch(self, opts):
        """[--ssh|--http] <user> [<refspec>]
           Fetch refs from a user's fork"""
        for fork in self.repository(opts).forks():
            if fork.owner['nickname'] in opts['<user>']:
                url = self.clone_url(fork, opts)
                refspec = opts['<refspec>'] or 'refs/heads/*'
                if ':' not in refspec:
                    if not refspec.startswith('refs/'):
                        refspec += ':' + 'refs/remotes/%s/' % fork.owner['nickname'] + refspec
                    else:
                        refspec += ':' + refspec.replace('refs/heads/', 'refs/remotes/%s/' % fork.owner['nickname'])
                self.gitm('fetch', url, refspec, redirect=False)

    @command
    def fork(self, opts):
        """[--ssh|--http] [--triangular [--upstream-branch=<branch>]] [<repo>]
           Fork a repo and clone it"""
        do_clone = bool(opts['<repo>'])
        repo = self.repository(opts)
        if repo.owner['nickname'] == self.my_login:
            err("You cannot fork your own repos")

        try:
            self.me.repository(repo.name)
            err("Repository already exists")
        except bbapi.BitBucketError:
            pass

        my_fork = repo.fork()
        self.wait_for_repo(my_fork.owner['nickname'], my_fork.name, opts)

        if do_clone:
            self.clone(opts, repo=my_fork)
        else:
            self.set_origin(opts, repo=my_fork)

    @command
    @wants_parent
    def forks(self, opts):
        """[<repo>]
           List all forks of this repository"""
        repo = self.repository(opts)
        print("[%s] %s" % (wrap(repo.owner['nickname'], attr.bright), repo.links['html']['href']))
        for fork in repo.forks():
            print("[%s] %s" % (fork.owner['nickname'], fork.links['html']['href']))

    @command
    def issue(self, opts):
        """[<repo>] [--parent] [<issue>...]
           Show issue details or report an issue"""
        if opts['<repo>'] and opts['<repo>'].isdigit():
            # Let's assume it's an issue
            opts['<issue>'].insert(0, opts['<repo>'])
            opts['<repo>'] = None
        repo = self.repository(opts)
        for issue_no in opts['<issue>']:
            try:
                issue = repo.issue(issue_no)
                print(wrap(issue.title.encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding), attr.bright, attr.underline))
                print(issue.content['raw'].encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding))
                print(issue.html_url)
            except bbapi.BitBucketError:
                bbe = sys.exc_info()[1]
                if bbe.args[0] == 'No Issue matches the given query.':
                    print('No issue with id %s found in repository %s' % (issue_no, repo.full_name))
                else:
                    raise
        if not opts['<issue>']:
            extra = """Reporting an issue on %s
Please describe the issue as clearly as possible. Lines starting with '#' will
be ignored, the first line will be used as title for the issue.""" % repo.full_name
            title, body = self.edit_msg(None, '', extra, 'ISSUE_EDITMSG')
            if not body:
                err("Empty issue message")
            try:
                issue = repo.create_issue(title=title, body=body)
                print("Issue %d created %s" % (issue.id, issue.html_url))
            except:
                filename = self.backup_message(title, body, 'issue-message-')
                err("Failed to create an issue, the issue text has been saved in %s" % filename)

    @command
    def issues(self, opts):
        """[<repo>] [--parent] [<filter>...]
           List issues in a repository"""
        if opts['<repo>'] and '=' in opts['<repo>']:
            opts['<filter>'].insert(0, opts['<repo>'])
            opts['<repo>'] = None
        if (not opts['<repo>'] and not self.in_repo) or opts['<repo>'] == '--':
            repos = self.me.repositories()
        else:
            repos = [self.repository(opts)]
        for repo in repos:
            if repo.fork and opts['--parent']:
                repo = self.parent_repo(repo) or repo
            filters = dict([x.split('=', 1) for x in opts['<filter>']])
            try:
                issues = repo.issues(**filters)
            except bbapi.BitBucketError:
                issues = None
            try:
                pullrequests = repo.pull_requests()
            except bbapi.BitBucketError:
                pullrequests = None

            if issues:
                print(wrap("Issues for %s" % repo.full_name, attr.bright))
                for issue in issues:
                    print("[%d] %s %s" % (issue.id, issue.title.encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding), issue.html_url))
            if pullrequests:
                print(wrap("Pull requests for %s" % repo.full_name, attr.bright))
                for pr in pullrequests:
                    print("[%d] %s %s" % (pr.id, pr.title.encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding), pr.html_url))


    @command
    def ls(self, opts):
        """[<dir>...]
           Display the contents of a directory on BitBucket"""
        for arg in opts['<dir>'] or ['']:
            repo, ref, file = ([None, None] + arg.split(':',2))[-3:]
            user = None
            if repo:
                user, repo = ([None] + repo.split('/'))[-2:]
                repo = self.bb.repository(user or self.my_login, repo)
            else:
                repo = self.repository(opts)
                file = self.rel2root(file)
            try:
                content = repo.src(path=file or '/', revision=ref or 'master') # BitBucket has no API to retrieve the default branch
            except bbapi.BitBucketError:
                err("No such file: %s" % arg)
            if isinstance(content, bytes):
                err("Not a directory: %s" % arg)

            ms = max([len(str(file.get('size', 0))) for file in content])
            content.sort(key=lambda x: x['path'])
            for entry in content:
                print("%*s %s" % (ms, entry.get('size', ''), entry['path']))

    @command
    def mirror(self, opts):
        """[--ssh|--http] [<repo>]
           Mirror a repository, or all repositories for a user"""
        if opts['<repo>'] and opts['<repo>'].endswith('/*'):
            user = opts['<repo>'].rsplit('/', 2)[-2]
            for repo in self.bb.user(user).repositories():
                opts['<repo>'] = repo.full_name
                self.mirror(opts)
            return
        repo = self.repository(opts)
        git_dir = repo.name + '.git'
        cur_dir = os.path.basename(os.path.abspath(os.getcwd()))
        if cur_dir != git_dir and not os.path.exists(git_dir):
            url = self.clone_url(repo, opts)
            self.gitm('clone', '--mirror', url, git_dir, redirect=False)
        else:
            if git_dir == cur_dir:
                git_dir = '.'
            # Update the current, mirrored repo
            if self.git('--git-dir', git_dir, 'config', 'core.bare').stdout.strip() != 'true' or \
               self.git('--git-dir', git_dir, 'config', 'remote.origin.mirror').stdout.strip() != 'true':
                   err("This is not a mirrored repository")
            self.gitm('--git-dir', git_dir, 'fetch', '-q', '--prune', 'origin', redirect=False)

        with open(os.path.join(git_dir, 'description'), 'w') as fd:
            fd.write(repo.description or "")

    @command
    def permissions(self, opts):
        """[<repo>]
           List repo permissions"""
        repo = self.repository(opts)
        order = {'admin': 0, 'write': 1, 'read': 2}
        permissions = repo.permissions()
        if not permissions:
            return
        permissions.sort(key=lambda permission: (order[permission['permission']], permission['user']['nickname']))
        maxlen = max([len(permission['user']['nickname']) for permission in permissions])
        fmt = "%%s %%-%ds (%%s)" % maxlen
        for permission in permissions:
            print(fmt % (wrap("%-5s" % permission['permission'], attr.faint), permission['user']['nickname'], permission['user']['display_name']))

    @command
    def public_keys(self, opts):
        """[<user>]
           Lists all keys for a user"""
        user = opts['<user>'] and self.bb.user(opts['<user>'][0]) or self.me
        for key in user.keys():
            print("%s %s" % (key.key, key.label or ''))

    @command
    def pull_request(self, opts):
        """[--yes] [<yours:theirs>]
           Opens a pull request to merge your branch to an upstream branch"""
        repo = self.repository(opts)
        if repo.parent:
            parent = self.parent_repo(repo)
        else:
            parent = repo
        # Which branch?
        src = opts['<yours:theirs>'] or ''
        dst = None
        if ':' in src:
            src, dst = src.split(':', 1)
        if not src:
            src = self.gitm('rev-parse', '--abbrev-ref', 'HEAD').stdout.strip()
        if not dst:
            dst = parent.mainbranch['name']
            tracking_branch = self.git('rev-parse', '--symbolic-full-name', '%s@{u}' % src).stdout.strip()
            if tracking_branch.startswith('refs/remotes/'):
                tracking_remote, tracking_branch = tracking_branch.split('/', 3)[-2:]
                if tracking_branch != src or repo.remote != tracking_remote:
                    # Interesting. We're not just tracking a branch in our clone!
                    dst = tracking_branch

        if src == dst and parent == repo:
            err("Cannot file a pull request on the same branch")

        # Try to get the local commit
        commit = self.gitm('show-ref', 'refs/heads/%s' % src).stdout.split()[0]
        # Do they exist on bitbucket?
        srcb = repo.branches().get(src, None)
        if not srcb:
            if self.question("Branch %s does not exist in your BitBucket repo, shall I push?" % src):
                self.gitm('push', '-u', repo.remote, src, redirect=False)
                srcb = repo.branches().get(src, None)
            else:
                err("Aborting")
        elif srcb and srcb.target['hash'] != commit:
            # Have we diverged? Then there are commits that are reachable from the bitbucket branch but not local
            diverged = self.gitm('rev-list', srcb.raw_node, '^' + commit)
            if diverged.stderr or diverged.stdout:
                if self.question("Branch %s has diverged from bitbucket, shall I push and overwrite?" % src, default=False):
                    self.gitm('push', '--force', repo.remote, src, redirect=False)
                else:
                    err("Aborting")
            else:
                if self.question("Branch %s not up to date on bitbucket, but can be fast forwarded, shall I push?" % src):
                    self.gitm('push', repo.remote, src, redirect=False)
                else:
                    err("Aborting")
            srcb = repo.branches().get(src, None)

        dstb = parent.branches().get(dst, None)
        if not dstb:
            err("Branch %s does not exist in %s/%s" % (dst, parent.owner.login, parent.name))

        # Do we have the dst locally?
        for remote in self.gitm('remote').stdout.strip().split("\n"):
            url = self.gitm('config', 'remote.%s.url' % remote).stdout.strip()
            if url in parent.links['clone'].values():
                if parent.is_private and url != parent.links['clone']['ssh']:
                    err("You should configure %s to fetch via ssh, it is a private repo" % parent.full_name)
                self.gitm('fetch', remote, redirect=False)
                break
        else:
            err("You don't have %ss configured as a remote repository" % parent.full_name)

        # How many commits?
        accept_empty_body = False
        commits = self.gitm('log', '--pretty=%H', '%s/%s..%s' % (remote, dst, src)).stdout.strip().split()
        commits.reverse()
        if not commits:
            err("Your branch has no commits yet")

        # 1 commit: title/body from commit
        if len(commits) == 1:
            title, body = self.gitm('log', '--pretty=%s\n%b', '%s^..%s' % (commits[0], commits[0])).stdout.split('\n', 1)
            title = title.strip()
            body = body.strip()
            accept_empty_body = not bool(body)

        # More commits: title from branchname (titlecased, s/-/ /g), body comments from shortlog
        else:
            title = src
            if '/' in title:
                title = title[title.rfind('/') + 1:]
            title = title.title().replace('-', ' ')
            body = ""

        extra = """Requesting a pull from %s/%s into %s/%s

Please enter a message to accompany your pull request. Lines starting
with '#' will be ignored, and an empty message aborts the request.""" % (repo.owner['nickname'], src, parent.owner['nickname'], dst)
        extra += "\n\n" + self.gitm('shortlog', '%s/%s..%s' % (remote, dst, src)).stdout.strip()
        extra += "\n\n" + self.gitm('diff', '--stat', '%s^..%s' % (commits[0], commits[-1])).stdout.strip()
        title, body = self.edit_msg(title, body, extra, 'PULL_REQUEST_EDIT_MSG')
        if not body and not accept_empty_body:
            err("No pull request message specified")

        try:
            pull = parent.create_pull_request(src=srcb, dst=dstb, title=title, body=body)
            print("Pull request %d created %s" % (pull.id, pull.links['html']['href']))
        except:
            filename = self.backup_message(title, body, 'pull-request-message-')
            err("Failed to create a pull request, the pull request text has been saved in %s" % filename)

    @command
    def remove_deploy_key(self, opts):
        """<key>...
           Remove deploy key by id"""
        repo = self.repository(opts)
        for key in opts['<key>']:
            repo.remove_deploy_key(key)

    @command
    def repos(self, opts):
        """[--no-forks] [<workspace>]
           List all repos of a workspace, or your repos"""
        if opts['<workspace>']:
            repos = self.bb.workspace(opts['<workspace>']).repositories()
        else:
            repos = self.me.repositories()
        if not repos:
            return
        maxlen = max([len(x.name) for x in repos])
        fmt = u"%%-%ds %%5s %%s" % maxlen
        for repo in repos:
            color = [attr.normal]
            if repo.is_private:
                color.append(fgcolor.red)
            if 'parent' in repo.data:
                if opts['--no-forks']:
                    continue
                color.append(attr.faint)
            print(wrap(fmt % (repo.full_name, '(%s)' % repo.scm, repo.description), *color))

    @command
    def set_origin(self, opts, repo=None, remote='origin'):
        """[--ssh|--http] [--triangular [--upstream-branch=<branch>]]
           Set the remote 'origin' to bitbucket.
           If this is a fork, set the remote 'upstream' to the parent"""
        if not repo:
            repo = self.repository(opts)
            # Is this mine? No? Do I have a clone?
            if repo.owner['nickname'] != self.my_login:
                try:
                    repo = self.me.repository(repo.slug)
                except bbapi.BitBucketError:
                    pass

        url = self.clone_url(repo, opts)
        if self.git('config', 'remote.%s.url' % remote).stdout.strip() != url:
            print("Pointing %s to %s" % (remote, url))
            self.gitm('config', 'remote.%s.url' % remote, url)
        self.gitm('config', '--replace-all', 'remote.%s.fetch' % remote, '+refs/heads/*:refs/remotes/%s/*' % remote)

        if repo.parent:
            parent = self.parent_repo(repo)
            url = self.clone_url(parent, opts)
            if self.git('config', 'remote.upstream.url').stdout.strip() != url:
                print("Pointing upstream to %s" % url)
                self.gitm('config', 'remote.upstream.url', url)
            self.gitm('config', 'remote.upstream.fetch', '+refs/heads/*:refs/remotes/upstream/*')

        if self.git('ls-remote', remote).stdout.strip():
            self.gitm('fetch', remote, redirect=False)
        if repo.parent:
            self.gitm('fetch', 'upstream', redirect=False)

        if remote != 'origin':
            return

        self.set_tracking_branches(remote, upstream="upstream", triangular=opts['--triangular'], upstream_branch=opts['--upstream-branch'])

    @command
    def snippet(self, opts):
        """[--description=<description>] <file>...
           Create a new snippet from files or stdin"""
        files = {}
        description = opts['--description'] or ''
        for f in opts['<file>']:
            if f == '-':
                files['stdout'] = sys.stdin.read()
            else:
                if not os.path.exists(f):
                    err("No such file: %s" % f)
                with open(f) as fd:
                    files[os.path.basename(f)] = fd.read()
        snippet = self.me.create_snippet(description=description, files=files)
        print("Snippet created at %s" % snippet.links['html']['href'])

    @command
    def snippets(self, opts):
        """[<user>]
           Show all snippets for a user"""
        snippets = self.bb.user(opts['<user>'] or self.my_login).snippets()
        for snippet in snippets:
            print("%s - %s" % (snippet.title, snippet.links['html']['href']))

    @command
    def whoami(self, opts):
        """\nDisplay BitBucket user info"""
        opts['<user>'] = [self.my_login]
        self.whois(opts)

    @command
    def whois(self, opts):
        """<user>...
           Display Bitbucket user info"""
        for user_ in opts['<user>']:
            user = self.bb.user(user_)
            if not user:
                print("No such user: %s" % user_)
                continue
            print(wrap(user.display_name or user.nickname, attr.bright, attr.underline))
            print("UUID:     %s" % user.account_id)
            print("Profile:  %s" % user.links['html']['href'])
            if user.location:
                print("Location: %s" % user.location)
            try:
                keys = user.keys()
            except bbapi.BitBucketError:
                keys = []
            for pkey in keys:
                algo, key = pkey.key.split()[:2]
                algo = algo[4:].upper()
                if pkey.label:
                    print("%s key%s...%s (%s)" % (algo, ' ' * (6 - len(algo)), key[-10:], pkey.label))
                else:
                    print("%s key%s...%s" % (algo, ' ' * (6 - len(algo)), key[-10:]))



import json
from operator import attrgetter
import requests
import uritemplate

def check(resp):
    if resp.status_code not in (200, 201, 204):
        try:
            message = resp.json()['error']['message']
        except (KeyError, ValueError):
            message = resp.content

        if resp.status_code == 401:
            raise BitBucketAuthenticationError(message)
        raise BitBucketError(message)
    if not resp.content:
        return None
    if not resp.headers['Content-Type'].startswith('application/json'):
        return resp.content
    return resp.json()

class Bitbucket(object):
    def __init__(self, username, passwd):
        self.username = username
        self.passwd = passwd

    def user(self, username=None):
        return User(self, username=username)

    def team(self, username):
        return Team(self, username=username)

    def workspace(self, workspace):
        return Workspace(self, workspace=workspace)

class BitBucketError(Exception):
    pass

class BitBucketAuthenticationError(BitBucketError):
    pass

class BBobject(object):
    def __new__(cls, *args, **kwargs):
        self = super(BBobject, cls).__new__(cls)
        return self

    def __init__(self, bb, mode='fetch', **kwargs):
        self.bb = bb
        self.data = {}
        if mode == 'fetch':
            for arg in kwargs:
                setattr(self, arg, kwargs[arg])
            self.url = kwargs.get('url', uritemplate.expand(self.uri, **kwargs)).replace('!api', 'api')
            self.data = self.get(self.url)
        elif mode == 'list':
            self.url = kwargs.get('url', uritemplate.expand(self.list_uri, **kwargs)).replace('!api', 'api')
            self.instances = []
            # FIXME properly handle pagination
            for instance in self.get(self.url)["values"]:
                kw = kwargs.copy()
                kw.update(instance)
                instance = type(self)(self.bb, mode=None, **kw)
                instance.url = instance.links['self']['href']
                self.instances.append(instance)
        else:
            self.data = kwargs
        for datum in self.data:
            if datum == 'data':
                setattr(self, '_' + datum, self.data[datum])
            else:
                setattr(self, datum, self.data[datum])

    @classmethod
    def list(klass, bb, **kwargs):
        return klass(bb, mode="list", **kwargs).instances

    def get(self, *args, **kwargs):
        kwargs.update({'auth': (self.bb.username, self.bb.passwd)})
        return check(requests.get(*args, **kwargs))

    def post(self, *args, **kwargs):
        kwargs.update({'auth': (self.bb.username, self.bb.passwd)})
        return check(requests.post(*args, **kwargs))

    def put(self, *args, **kwargs):
        kwargs.update({'auth': (self.bb.username, self.bb.passwd)})
        return check(requests.put(*args, **kwargs))

    def delete_(self, *args, **kwargs):
        kwargs.update({'auth': (self.bb.username, self.bb.passwd)})
        return check(requests.delete(*args, **kwargs))

class User(BBobject):
    uri = 'https://bitbucket.org/api/2.0/users/{id}'

    def __init__(self, bb, username):
        team = bb.team(username)
        super(User, self).__init__(bb, id=team.account_id)

    def __eq__(self, other):
        if isinstance(other, str):
            return other == self.account_id
        if isinstance(other, dict):
            return other.get('account_id', None) == self.account_id
        return super(User, self).__eq__(other)

    def repository(self, slug):
        return Repository(self.bb, owner=self.nickname, slug=slug)

    def create_repository(self, slug, description, is_private, has_issues, has_wiki):
        data = {'owner': self.nickname, 'slug': slug, 'description': description, 'is_private': is_private,
                'scm': 'git', 'has_issues': has_issues, 'has_wiki': has_wiki}
        repo = self.post(uritemplate.expand(Repository.uri, slug=slug, workspace=self.nickname), data=json.dumps(data), headers={'content-type': 'application/json'})
        repo['is_fork'] = False
        return Repository(self.bb, mode=None, **repo)

    def create_key(self, key, label):
        url = uritemplate.expand(Key.uri, user=self.account_id)
        data = self.post(url, data={'user': self.account_id, 'key': key, 'label': label})
        return Key(self.bb, mode=None, **data)

    def keys(self):
        return Key.list(self.bb, user=self.account_id)

    def repositories(self):
        url = 'https://bitbucket.org/api/2.0/repositories?role=member'
        data = self.get(url)['values']
        return [Repository(self.bb, mode=None, **entry) for entry in data]

    def snippets(self):
        url = 'https://bitbucket.org/api/2.0/snippets/'
        data = self.get(url)['values']
        return [Snippet(self.bb, mode=None, **snippet) for snippet in data]

    def create_snippet(self, description, files):
        url = 'https://bitbucket.org/api/2.0/snippets/'
        data = {'scm': 'git', 'is_private': 'false', 'title': description}
        files = [('file', (filename, content)) for (filename, content) in files.items()]
        snippet = self.post(url, data=data, files=files)
        return Snippet(self.bb, mode=None, **snippet)

class Team(BBobject):
    uri = 'https://bitbucket.org/api/2.0/teams/{username}'

def ssh_fix(url):
    if not url.startswith('ssh://'):
        return url
    return ':'.join(url[6:].split('/',1))

class PullRequest(BBobject):
    uri = 'https://api.bitbucket.org/2.0/repositories/{owner}/{slug}/pullrequests/{id}'

    def get_url(self):
        return self.links['html']['href']

    html_url = property(get_url)

class Repository(BBobject):
    uri = 'https://bitbucket.org/api/2.0/repositories/{workspace}/{slug}'
    list_uri = 'https://bitbucket.org/api/2.0/repositories/{workspace}/'

    def __init__(self, *args, **kwargs):
        self.parent = None
        super(Repository, self).__init__(*args, **kwargs)
        if not hasattr(self, 'links'):
            return
        links, self.links['clone'] = self.links['clone'], {}
        for link in links:
            self.links['clone'][link['name']] = ssh_fix(link['href'])


    def fork(self):
        self.post(self.url + '/forks', data={'name': self.name})
        for _ in range(5):
            try:
                return self.bb.workspace(self.bb.username).repository(self.name)
            except BitBucketError:
                print("Waiting for repository to be forked...")

    def branches(self):
        branches = self.get(self.url + '/refs')['values']
        return dict([(branch['name'], Branch(self.bb, mode=None, repository=self, **branch)) for branch in branches if branch['type'] == 'branch'])

    def pull_requests(self, **params):
        url = 'https://bitbucket.org/api/2.0/repositories/%s/pullrequests?state=OPEN' % self.full_name
        data = self.get(url)['values']
        return [PullRequest(self.bb, mode=None, **pr) for pr in data]

    def pull_request(self, number):
        owner, slug = self.full_name.split('/')
        return PullRequest(self.bb, owner=owner, slug=slug, id=number)

    def create_pull_request(self, src, dst, title, body):
        data = {'title': title, 'description': body}
        data['source'] = {'branch': {'name': src.name}, 'repository': {'full_name': src.repository.full_name}}
        data['destination'] = {'branch': {'name': dst.name}}
        pr = self.post(self.url + '/pullrequests', data=json.dumps(data), headers={'content-type': 'application/json'})
        return PullRequest(self.bb, mode=None, **pr)

    def forks(self):
        data = self.get(self.links['forks']['href'].replace('!api', 'api'))['values']
        return [Repository(self.bb, mode=None, **repo) for repo in data]

    def issues(self, **params):
        url = 'https://bitbucket.org/api/2.0/repositories/%s/issues' % self.full_name
        data = self.get(url, data=params)['values']
        return [Issue(self.bb, mode=None, **issue) for issue in data]

    def issue(self, id):
        return Issue(self.bb, owner=self.owner['nickname'], slug=self.slug, id=id, repo=self)

    def create_issue(self, title, body):
        data = {'title': title, 'content': body}
        issue = self.post(self.url + '/issues', data=json.dumps(data))
        issue['repo'] = self
        return Issue(self.bb, owner=self.owner['nickname'], slug=self.slug, id=issue['local_id'], repo=self)

    def src(self, revision, path):
        uri = 'https://bitbucket.org/api/2.0/repositories/{workspace}/{slug}/src/{node}/{path}'
        url = uritemplate.expand(uri, workspace=self.owner['nickname'], slug=self.slug, node=revision, path=path)
        data = self.get(url)
        if isinstance(data, bytes):
            return data
        return data['values']

    def delete(self):
        if not hasattr(self, 'url'):
            data = {'owner': self.owner['account_id'], 'slug': self.name}
            self.url = [uritemplate.expand(x, **data) for x in self.uri]
        return self.delete_(self.url[1])

    def permissions(self):
        data = {'workspace': self.owner['nickname'], 'slug': self.name}
        url = uritemplate.expand('https://bitbucket.org/api/2.0/workspaces/{workspace}/permissions/repositories/{slug}', data)
        return self.get(url)['values']

    def add_deploy_key(self, key, label):
        url = self.url + '/deploy-keys'
        return self.post(url, {'key': key, 'label': label})

    def remove_deploy_key(self, key):
        url = self.url + '/deploy-keys/' + key
        return self.delete_(url)

    def deploy_keys(self):
        url = self.url + '/deploy-keys'
        return self.get(url)['values']

class Branch(BBobject):
    pass

class Key(BBobject):
    uri = 'https://bitbucket.org/api/2.0/users/{user}/ssh-keys/{key_id}'
    list_uri = 'https://bitbucket.org/api/2.0/users/{user}/ssh-keys'

    def delete(self):
        if not hasattr(self, 'url'):
            self.url = [uritemplate.expand(x, user=self.owner['username']) for x in self.uri]
        self.delete_(self.url[0] + '/%d' % self.pk)

class Issue(BBobject):
    uri = 'https://bitbucket.org/api/2.0/repositories/{owner}/{slug}/issues/{id}'

    def get_url(self):
        return self.links['html']['href']

    html_url = property(get_url)

class Snippet(BBobject):
    uri = 'https://bitbucket.org/api/2.0/snippets/{owner}/{id}'

    def __init__(self, *args, **kwargs):
        super(Snippet, self).__init__(*args, **kwargs)
        if not hasattr(self, 'links'):
            return
        links, self.links['clone'] = self.links['clone'], {}
        for link in links:
            self.links['clone'][link['name']] = ssh_fix(link['href'])

    def delete(self):
        if not hasattr(self, 'url'):
            data = {'owner': self.owner['username'], 'id': self.id}
            self.url = [uritemplate.expand(x, **data) for x in self.uri]
        return self.delete_(self.url[0])

class Workspace(BBobject):
    uri = 'https://bitbucket.org/api/2.0/workspaces/{workspace}'
    repositories_uri = 'https://bitbucket.org/api/2.0/repositories?role=member'

    def repositories(self):
        return Repository.list(self.bb, workspace=self.slug)

    def repository(self, slug):
        return Repository(self.bb, workspace=self.slug, slug=slug)




import sys

class Attr(object):
    def __init__(self, **attr):
        self.attr = attr
        self.rev_attr = dict([(v,k) for k,v in attr.items()])
        for k, v in attr.items():
            setattr(self, k, v)

    def name(self, val):
        return self.rev_attr[val]

    def xterm(self, val):
        return '%d;5;%d' % (self._xterm, val)

fgcolor = Attr(black=30, red=31, green=32, yellow=33, blue=34, magenta=35, cyan=36, white=37, _xterm=38, default=39, none=None)
bgcolor = Attr(black=40, red=41, green=42, yellow=43, blue=44, magenta=45, cyan=46, white=47, _xterm=48, default=49, none=None)
attr    = Attr(normal=0, bright=1, faint=2, underline=4, negative=7, conceal=8, crossed=9, none=None)

esc = '\033'
mode = lambda *args: "%s[%sm" % (esc, ';'.join([str(x) for x in args if x is not None]))
reset = mode(attr.normal)
if sys.stdout.isatty():
    wrap = lambda text, *args: "%s%s%s" % (mode(*args), text, reset)
else:
    wrap = lambda text, *args: text

erase_line = esc + '[K'
erase_display = esc + '[2J'
save_cursor = esc + '[s'
restore_cursor = esc + '[u'
cursor_up = lambda n: esc + '[%dA' % n

del sys
del Attr



import re
import glob
import os

escape_code_re = re.compile('\033\\[([^m]+)m')
codes = {
    1: {"font-weight": "bold"},
    4: {"text-decoration": "underline"},
    30: {"color": "black"},
    31: {"color": "red"},
    32: {"color": "green"},
    33: {"color": "yellow"},
    34: {"color": "blue"},
    35: {"color": "magenta"},
    36: {"color": "cyan"},
    37: {"color": "white"},
    40: {"background_color": "black"},
    41: {"background_color": "red"},
    42: {"background_color": "green"},
    43: {"background_color": "yellow"},
    44: {"background_color": "blue"},
    45: {"background_color": "magenta"},
    46: {"background_color": "cyan"},
    47: {"background_color": "white"},
}

def process_escape_html(match):
    items = [int(x) for x in match.group(1).split(';')]
    if 0 in items:
        return "</span>"
    attrs = [codes[x] for x in codes if x in items]
    if 38 in items:
        color = items[items.index(38)+2]
        color = {
            22:  "005f00",
            28:  "008700",
            64:  "5f8700",
            65:  "5f875f",
            237: "3a3a3a",
        }[color]
        return '<span style="color: #%s">' % color
    return '<span style="%s">' % '; '.join([", ".join([': '.join(x) for x in attr.items()]) for attr in attrs])

def mangle_html(app, exception):
    if app.builder.name != 'html' or exception:
        return
    for file in glob.glob(os.path.join(app.builder.outdir, '*.html')):
        with open(file, 'r+', encoding='utf-8') as fd:
            data = fd.read()
            data2 = re.sub(escape_code_re, process_escape_html, data)
            if data2 != data:
                fd.seek(0, os.SEEK_SET)
                fd.truncate()
                fd.write(data2)

def process_escape_man(match):
    items = [int(x) for x in match.group(1).split(';')]
    if 0 in items:
        return "\\fP"
    if 1 in items:
        return '\\fB'
    if 4 in items:
        return '\\fI'
    return ''

def mangle_man(app, exception):
    if app.builder.name != 'man' or exception:
        return
    for file in glob.glob(os.path.join(app.builder.outdir, '*.1')):
        with open(file, 'r+') as fd:
            data = fd.read()
            data2 = re.sub(escape_code_re, process_escape_man, data)
            if data2 != data:
                fd.seek(0, os.SEEK_SET)
                fd.truncate()
                fd.write(data2)

def setup(app):
    app.connect('build-finished', mangle_html)
    app.connect('build-finished', mangle_man)



from __future__ import absolute_import
from gitspindle import *
from gitspindle.ansi import *
import gitlab
import base64
import datetime
import getpass
import glob
import json
import os
import requests
import sys
import time
import webbrowser

class GitLab(GitSpindle):
    prog = 'git lab'
    what = 'GitLab'
    spindle = 'gitlab'
    hosts = ['gitlab.com', 'www.gitlab.com']
    api = gitlab
    access_levels = {
        'guest':     10,
        'reporter':  20,
        'developer': 30,
        'master':    40,
        'owner':     50,
    }
    access_levels_r = dict([(value, key) for (key, value) in access_levels.items()])

    # Support functions
    def login(self):
        self.gl = None
        host = self.config('host') or 'https://gitlab.com'
        if not host.startswith(('http://', 'https://')):
            try:
                requests.get('https://' + host)
            except:
                err("%s is not reachable via https. Use http://%s to use the insecure http protocol" % (host, host))
            host = 'https://' + host
        self.host = host

        user = self.config('user')
        if not user:
            user = input("GitLab user: ").strip()
            self.config('user', user)

        token = self.config('token')
        if not token:
            print("Generate a personal access token at %s/-/profile/personal_access_tokens and enter it below" % host)
            token = getpass.getpass("GitLab personal access token: ")
            self.gl = gitlab.Gitlab(host, private_token=token)
            self.gl.auth()
            self.config('token', token)
            location = '%s - do not share this file' % self.config_file
            if self.use_credential_helper:
                location = 'git\'s credential helper'
            print("Your GitLab authentication token is now stored in %s" % location)

        if not user or not token:
            err("No user or token specified")

        if not self.gl:
            self.gl = gitlab.Gitlab(host, private_token=token)
            try:
                self.gl.auth()
            except gitlab.GitlabAuthenticationError:
                # Token obsolete
                self.config('token', None)
                self.login()
        self.me = self.gl.user
        self.my_login = self.me.username

    def parse_url(self, url):
        return ([self.my_login] + url.path.rsplit('/',1))[-2:]

    def get_repo(self, remote, user, repo):
        if remote:
            id = self.git('config', 'remote.%s.gitlab-id' % remote).stdout.strip()
            if id and id.isdigit():
                return self.gl.projects.get(id)

        try:
            repo_ = self.gl.projects.get('%s/%s' % (user, repo))
        except gitlab.exceptions.GitlabGetError:
            return None
        if remote:
            self.gitm('config', 'remote.%s.gitlab-id' % remote, repo_.id)
        return repo_

    def clone_url(self, repo, opts):
        if opts['--ssh'] or not repo.visibility == 'public':
            return repo.ssh_url_to_repo
        if opts['--http']:
            return repo.http_url_to_repo
        if repo.namespace['full_path'] == self.my_login:
            return repo.ssh_url_to_repo
        return repo.http_url_to_repo

    def parent_repo(self, repo):
       if getattr(repo, 'forked_from_project', False):
           return self.gl.projects.get(repo.forked_from_project['id'])

    # There's no way to fetch a group by name. Abuse search.
    def find_group(self, name):
        try:
            for group in self.gl.Group(search=name):
                if group.path == name:
                    return group
        except gitlab.GitlabListError:
            pass

    def api_root(self):
        if hasattr(self, 'gl') and self.gl:
            return self.gl._url
        host = self.config('host')
        if not host:
            return 'https://gitlab.com/api/v3'
        host = host.rstrip('/') + '/api/v3'
        if not host.startswith(('http://', 'https://')):
            host = 'https://' + host
        return host

    # commands

    @command
    def add_public_keys(self, opts):
        """[<key>...]
           Adds keys to your public keys"""
        if not opts['<key>']:
            opts['<key>'] = glob.glob(os.path.join(os.path.expanduser('~'), '.ssh', 'id_*.pub'))
        existing = [x.key.split()[1] for x in self.me.keys.list()]
        for arg in opts['<key>']:
            with open(arg) as fd:
                algo, key, title = fd.read().strip().split(None, 2)
            if key in existing:
                continue
            key = "%s %s" % (algo, key)
            print("Adding %s" % arg)
            self.me.keys.create({'title': title, 'key': key})

    @command
    def add_member(self, opts):
        """[--access-level=guest|reporter|developer|master|owner] <user>...
           Add a project member"""
        repo = self.repository(opts)
        for user in opts['<user>']:
            if not isinstance(user, (gitlab.v4.objects.User, gitlab.v4.objects.CurrentUser)):
                user_ = self.gl.users.list(username=user)
                if not user_:
                    print("No such user: %s" % user)
                    continue
                user = user_[0]
            access_level = self.access_levels[opts['--access-level'] or 'developer']
            repo.members.create({"user_id": user.id, "access_level": access_level})

    @command
    def add_remote(self, opts):
        """[--ssh|--http] <user> [<name>]
           Add user's fork as a named remote. The name defaults to the user's loginname"""
        dwim = self.repository(opts)
        user = opts['<user>'][0]
        name = opts['<name>'] or user
        try:
            repo = self.gl.projects.get('%s/%s' % (user, dwim.path))
        except gitlab.exceptions.GitlabGetError:
            err("Repository %s/%s does not exist" % (user, dwim.path))
        url = self.clone_url(repo, opts)
        self.gitm('remote', 'add', name, url)
        self.gitm('config', 'remote.%s.gitlab-id' % name, repo.id)
        self.gitm('fetch', name, redirect=False)

    @command
    def apply_merge(self, opts):
        """[--ssh|--http] <merge-request-number>
           Applies a merge request as a series of cherry-picks"""
        repo = self.repository(opts)
        mn = int(opts['<merge-request-number>'])
        for req in repo.mergerequests.list():
            if req.iid == mn:
                mr = req
                break
        else:
            err("Merge request %s does not exist" % opts['<merge-request-number>'])
        print("Applying merge request #%d from %s: %s" % (mr.iid, mr.author['name'], mr.title))
        # Warnings
        warned = False
        cbr = self.gitm('rev-parse', '--symbolic-full-name', 'HEAD').stdout.strip().replace('refs/heads/','')
        if cbr != mr.target_branch:
            print(wrap("Merge request was filed against %s, but you're on the %s branch" % (mr.target_branch, cbr), fgcolor.red))
            warned = True
        if mr.state == 'merged':
            print(wrap("Merge request was already merged", fgcolor.red))
        if mr.state == 'closed':
            print(wrap("Merge request has already been closed", fgcolor.red))
            warned = True
        if warned:
            if not self.question("Continue?", default=False):
                sys.exit(1)
        # Fetch mr if needed
        sha = self.git('rev-parse', '--verify', 'refs/merge/%d/head' % mr.iid).stdout.strip()
        if not sha:
            print("Fetching merge request")
            url = self.clone_url(self.gl.projects.get(mr.source_project_id), opts)
            self.gitm('fetch', url, 'refs/heads/%s:refs/merge/%d/head' % (mr.source_branch, mr.iid), redirect=False)
        head_sha = self.gitm('rev-parse', 'HEAD').stdout.strip()
        if self.git('merge-base', 'refs/merge/%d/head' % mr.iid, head_sha).stdout.strip() == head_sha:
            print("Fast-forward merging %s..refs/merge/%d/head" % (mr.target_branch, mr.iid))
            self.gitm('merge', '--ff-only', 'refs/merge/%d/head' % mr.iid, redirect=False)
        else:
            print("Cherry-picking %s..refs/merge/%d/head" % (mr.target_branch, mr.iid))
            self.gitm('cherry-pick', '%s..refs/merge/%d/head' % (mr.target_branch, mr.iid), redirect=False)

    @command
    def browse(self, opts):
        """[--parent] [<repo>] [<section>]
           Open the GitLab page for a repository in a browser"""
        sections = ['issues', 'merge_requests', 'wiki', 'files', 'commits', 'branches', 'graphs', 'settings']
        if opts['<repo>'] in sections and not opts['<section>']:
            opts['<repo>'], opts['<section>'] = None, opts['<repo>']
        repo = self.repository(opts)
        section_map = {'wiki': 'wikis/home', 'files': 'tree/%s' % repo.default_branch,
                       'commits': 'commits/%s' % repo.default_branch, 'settings': 'edit',
                       'graphs': 'graphs/%s' % repo.default_branch}
        url = repo.web_url
        if opts['<section>']:
            url += '/' + section_map.get(opts['<section>'], opts['<section>'])
        webbrowser.open_new_tab(url)

    @command
    def calendar(self, opts):
        """[<user>]
           Show a timeline of a user's activity"""
        user = (opts['<user>'] or [self.my_login])[0]
        user = self.gl.users.list(username=user)[0]
        months = []
        rows = [[],[],[],[],[],[],[]]
        commits = []
        url = self.gl._url
        url = '%s/%s/%s/calendar' % (url[:url.find('/api')], 'users', user.username)
        data = requests.get(url).json()


        first = datetime.datetime.today() - datetime.timedelta(365)
        wd = (first.weekday()+1) % 7
        for i in range(wd):
            rows[i].append((None,None))
        if wd:
            months.append(first.month)
        for back in range(364, -1, -1):
            date = datetime.datetime.today() - datetime.timedelta(back)
            ts = date.strftime('%Y-%m-%d')
            wd = (date.weekday()+1) % 7
            count = data.get(ts, 0)
            rows[wd].append((date.day, count))
            if not wd:
                months.append(date.month)
            if count:
                commits.append(count)

        # Print months
        sys.stdout.write("  ")
        last = -1
        skip = months[2] != months[0]
        monthtext = ('', 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')
        for month in months:
            if month != last:
                sys.stdout.write(monthtext[month] + ' ')
                skip = True
                last = month
            elif not skip:
                sys.stdout.write('  ')
            else:
                skip = False
        print("")

        # Print commits
        days = 'SMTWTFS'
        commits.sort()
        if len(commits) < 2:
            p5 = p15 = p35 = 0
        else:
            p5  = commits[min(int(round(len(commits) * 0.95)), len(commits)-1)]
            p15 = commits[min(int(round(len(commits) * 0.85)), len(commits)-1)]
            p35 = commits[min(int(round(len(commits) * 0.65)), len(commits)-1)]
        blob1 = b'\xe2\x96\xa0'.decode('utf-8')
        blob2 = b'\xe2\x97\xbc'.decode('utf-8')
        for rnum, row in enumerate(rows):
            if rnum % 2:
                sys.stdout.write(days[rnum] + " ")
            else:
                sys.stdout.write("  ")
            for (day, count) in row:
                if count is None:
                    color = attr.conceal
                elif count > p5:
                    color = fgcolor.xterm(22)
                elif count > p15:
                    color = fgcolor.xterm(28)
                elif count > p35:
                    color = fgcolor.xterm(64)
                elif count:
                    color = fgcolor.xterm(65)
                else:
                    color = fgcolor.xterm(237)
                if day == 1:
                    msg = wrap(blob2, attr.underline, color)
                    sys.stdout.write(msg)
                else:
                    msg = wrap(blob1, color)
                    sys.stdout.write(msg)
                sys.stdout.write(' ')
            print("")


    @command
    def cat(self, opts):
        """<file>...
           Display the contents of a file on GitLab"""
        for file in opts['<file>']:
            repo, ref, file = ([None, None] + file.split(':',2))[-3:]
            user = None
            if repo:
                user, repo = ([None] + repo.rsplit('/', 1))[-2:]
                repo = self.gl.projects.get('%s/%s' % (user or self.my_login, repo))
            else:
                repo = self.repository(opts)
                file = self.rel2root(file).lstrip('/')

            try:
                file = repo.files.get(ref=ref or repo.default_branch, file_path=file)
                os.write(sys.stdout.fileno(), base64.b64decode(file.content))
            except gitlab.GitlabGetError:
                sys.stderr.write("No such file: %s\n" % file)

    @command
    def clone(self, opts, repo=None):
        """[--ssh|--http] [--triangular [--upstream-branch=<branch>]] [--parent] [git-clone-options] <repo> [<dir>]
           Clone a repository by name"""
        if not repo:
            repo = self.repository(opts)
        url = self.clone_url(repo, opts)

        args = opts['extra-opts']
        args.append(url)
        dir = opts['<dir>'] and opts['<dir>'][0] or repo.path
        if '--bare' in args:
            dir += '.git'
        args.append(dir)

        self.gitm('clone', *args, redirect=False).returncode
        self.gitm('config', 'remote.origin.gitlab-id', repo.id, cwd=dir)
        if hasattr(repo, 'forked_from_project'):
            os.chdir(dir)
            self.set_origin(opts, repo=repo)

    @command
    def create(self, opts):
        """[--private|--internal] [--group=<group>] [--description=<description>]
           Create a repository on gitlab to push to"""
        root = self.gitm('rev-parse', '--show-toplevel').stdout.strip()
        name = os.path.basename(root)
        if (opts['--group'] or self.my_login, name) in [(x.namespace['full_path'], x.path) for x in self.gl.projects.list()]:
            err("Repository already exists")
        visibility_level = 20 # public
        if opts['--internal']:
            visibility_level = 10
        elif opts['--private']:
            visibility_level = 0
        kwargs = {'name': name, 'description': opts['--description'] or "", 'visibility_level': visibility_level}
        if opts['--group']:
            group = self.find_group(opts['--group'])
            if not group:
                err("Group %s could not be found" % opts['--group'])
            kwargs['namespace_id'] = group.id
        i = 0
        while True:
            try:
                repo = self.gl.projects.create(kwargs)
                repo.save()
                break
            except gitlab.exceptions.GitlabCreateError as gce:
                i += 1
                time.sleep(1)
                if (gce.response_code != 400) \
                        or (not isinstance(gce.error_message, dict)) \
                        or (not 'base' in gce.error_message) \
                        or (not isinstance(gce.error_message['base'], list)) \
                        or (not 'The project is still being deleted. Please try again later.' in gce.error_message['base']) \
                        or (i >= 120):
                    raise
        if 'origin' in self.remotes():
            print("Remote 'origin' already exists, adding the GitLab repository as 'gitlab'")
            self.set_origin(opts, repo=repo, remote='gitlab')
        else:
            self.set_origin(opts, repo=repo)

    @command
    def fetch(self, opts):
        """[--ssh|--http] <user> [<refspec>]
           Fetch refs from a user's fork"""
        repo = self.repository(opts)
        user = opts['<user>'][0]
        refspec = opts['<refspec>'] or 'refs/heads/*'
        try:
            repo = self.gl.projects.get('%s/%s' % (user, repo.path))
        except gitlab.exceptions.GitlabGetException:
            err("Repository %s/%s does not exist" % (user, repo.path))
        if ':' not in refspec:
            if not refspec.startswith('refs/'):
                refspec += ':' + 'refs/remotes/%s/' % repo.namespace['full_path'] + refspec
            else:
                refspec += ':' + refspec.replace('refs/heads/', 'refs/remotes/%s/' % repo.namespace['full_path'])
        url = self.clone_url(repo, opts)
        self.gitm('fetch', url, refspec, redirect=False)

    @command
    def fork(self, opts):
        """[--ssh|--http] [--triangular [--upstream-branch=<branch>]] [<repo>]
           Fork a repo and clone it"""
        do_clone = bool(opts['<repo>'])
        repo = self.repository(opts)
        if repo.namespace['full_path'] == self.my_login:
            err("You cannot fork your own repos")

        try:
            self.gl.projects.get('%s/%s' % (self.my_login, repo.path))
            err("Repository already exists")
        except gitlab.exceptions.GitlabGetError:
            pass

        i = 0
        success = False
        while not success:
            try:
                my_fork = repo.forks.create({})
                success = True
            except gitlab.GitlabForkError as gfe:
                i += 1
                time.sleep(1)
                if (gfe.response_code != 409) \
                        or (not isinstance(gfe.error_message, dict)) \
                        or (not 'base' in gfe.error_message) \
                        or (not isinstance(gfe.error_message['base'], list)) \
                        or (not 'The project is still being deleted. Please try again later.' in gfe.error_message['base']) \
                        or (i >= 120):
                    raise

        self.wait_for_repo(my_fork.owner['username'], my_fork.name, opts)

        if do_clone:
            self.clone(opts, repo=my_fork)
        else:
            self.set_origin(opts, repo=my_fork)

    @command
    def issue(self, opts):
        """[<repo>] [--parent] [<issue>...]
           Show issue details or report an issue"""
        if opts['<repo>'] and opts['<repo>'].isdigit():
            # Let's assume it's an issue
            opts['<issue>'].insert(0, opts['<repo>'])
            opts['<repo>'] = None
        repo = self.repository(opts)
        for issue_no in opts['<issue>']:
            issues = repo.issues.list(iid=issue_no)
            if len(issues):
                issue = issues[0]
                print(wrap(issue.title.encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding), attr.bright, attr.underline))
                print(issue.description.encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding))
                print(issue.web_url)
            else:
                print('No issue with id %s found in repository %s' % (issue_no, repo.path_with_namespace))
        if not opts['<issue>']:
            extra = """Reporting an issue on %s/%s
Please describe the issue as clearly as possible. Lines starting with '#' will
be ignored, the first line will be used as title for the issue.""" % (repo.namespace['full_path'], repo.path)
            title, body = self.edit_msg(None, '', extra, 'ISSUE_EDITMSG')
            if not body:
                err("Empty issue message")

            try:
                issue = repo.issues.create({'title': title, 'description': body})
                issue.save()
                print("Issue %d created %s" % (issue.iid, issue.web_url))
            except:
                filename = self.backup_message(title, body, 'issue-message-')
                err("Failed to create an issue, the issue text has been saved in %s" % filename)
                raise

    @command
    def issues(self, opts):
        """[<repo>] [--parent] [<filter>...]
           List issues in a repository"""
        if opts['<repo>'] and '=' in opts['<repo>']:
            opts['<filter>'].insert(0, opts['<repo>'])
            opts['<repo>'] = None
        if (not opts['<repo>'] and not self.in_repo) or opts['<repo>'] == '--':
            repos = list(self.gl.projects.list())
        else:
            repos = [self.repository(opts)]
        for repo in repos:
            if opts['--parent']:
                repo = self.parent_repo(repo) or repo
            filters = dict([x.split('=', 1) for x in opts['<filter>']])
            issues = repo.issues.list(**filters)
            mergerequests = repo.mergerequests.list(state='opened')
            if not issues and not mergerequests:
                continue
            if issues:
                print(wrap("Issues for %s/%s" % (repo.namespace['full_path'], repo.path), attr.bright))
                for issue in issues:
                    print("[%d] %s %s" % (issue.iid, issue.title.encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding), issue.web_url))
            if mergerequests:
                print(wrap("Merge requests for %s/%s" % (repo.namespace['full_path'], repo.path), attr.bright))
                for mr in mergerequests:
                    print("[%d] %s %s" % (mr.iid, mr.title.encode(sys.stdout.encoding, errors='backslashreplace').decode(sys.stdout.encoding), mr.web_url))

    @command
    def log(self, opts):
        """[<repo>]
           Display GitLab log for a repository"""
        repo = self.repository(opts)
        if not repo:
            return
        now = datetime.datetime.now()
        for event in reversed(repo.events.list()):
            ts = datetime.datetime.strptime(event.created_at, '%Y-%m-%dT%H:%M:%S.%fZ')
            if ts.year == now.year:
                if (ts.month, ts.day) == (now.month, now.day):
                    ts = wrap(ts.strftime("%H:%M"), attr.faint)
                else:
                    ts = wrap(ts.strftime("%m/%d %H:%M"), attr.faint)
            else:
                ts = wrap(ts.strftime("%Y/%m/%d %H:%M"), attr.faint)
            if event.action_name in ('joined', 'left'):
                print('%s %s %s' % (ts, event.author_username, event.action_name))
            elif event.target_type == 'Issue':
                issue = repo.issues.get(event.target_iid)
                print('%s %s %s issue %s (%s)' % (ts, event.author_username, event.action_name, issue.iid, issue.title))
            elif event.target_type == 'MergeRequest':
                issue = repo.mergerequests.get(event.target_iid)
                print('%s %s %s merge request %s (%s)' % (ts, event.author_username, event.action_name, issue.iid, issue.title))
            elif event.target_type in ('Note', 'DiscussionNote'):
                print('%s %s created a comment' % (ts, event.author_username))
            elif event.action_name == 'pushed to':
                if event.push_data['commit_count'] == 0:
                    print('%s %s deleted branch %s' % (ts, event.author_username, event.push_data['ref']))
                else:
                    print('%s %s pushed %s commits to %s' % (ts, event.author_username, event.push_data['commit_count'], event.push_data['ref']))
            elif 'ref' in event.attributes:
                print('%s %s created tag %s' % (ts, event.author_username, event.attributes['ref'][10:]))
            else:
                print(wrap("Cannot display event. Please file a bug at github.com/seveas/git-spindle\nincluding the following output:", attr.bright))
                pprint(event.attributes)

    @command
    def ls(self, opts):
        """[<dir>...]
           Display the contents of a directory on GitLab"""
        for arg in opts['<dir>'] or ['']:
            repo, ref, file = ([None, None] + arg.split(':',2))[-3:]
            user = None
            if repo:
                user, repo = ([None] + repo.split('/'))[-2:]
                repo = self.gl.projects.get('%s/%s' % (user or self.my_login, repo))
            else:
                repo = self.repository(opts)
                file = self.rel2root(file).lstrip('/')

            try:
                content = repo.repository_tree(ref_name=ref or repo.default_branch, path=file)
            except gitlab.GitlabGetError:
                err("No such file: %s" % arg)
            if not content:
                err("Not a directory: %s" % arg)
            mt = max([len(file['type']) for file in content])
            fmt = "%%(mode)s %%(type)-%ds %%(id).7s %%(name)s" % (mt, )
            for file in content:
                print(fmt % file)

    @command
    def members(self, opts):
        """[<repo>]
           List repo memberships"""
        repo = self.repository(opts)
        members = repo.members.list()
        members.sort(key=lambda member: (-member.access_level, member.username))
        maxlen = members and max([len(member.username) for member in members]) or 0
        fmt = "%%s %%-%ds (%%s)" % maxlen
        for member in members:
            print(fmt % (wrap("%-9s" % self.access_levels_r[member.access_level], attr.faint), member.username, member.name))

    @command
    def merge_request(self, opts):
        """[--yes] [<yours:theirs>]
           Opens a merge request to merge your branch to an upstream branch"""
        repo = self.repository(opts)
        parent = self.parent_repo(repo) or repo
        # Which branch?
        src = opts['<yours:theirs>'] or ''
        dst = None
        if ':' in src:
            src, dst = src.split(':', 1)
        if not src:
            src = self.gitm('rev-parse', '--abbrev-ref', 'HEAD').stdout.strip()
        if not dst:
            dst = parent.default_branch
            tracking_branch = self.git('rev-parse', '--symbolic-full-name', '%s@{u}' % src).stdout.strip()
            if tracking_branch.startswith('refs/remotes/'):
                tracking_remote, tracking_branch = tracking_branch.split('/', 3)[-2:]
                if tracking_branch != src or repo.remote != tracking_remote:
                    # Interesting. We're not just tracking a branch in our clone!
                    dst = tracking_branch

        if src == dst and parent == repo:
            err("Cannot file a merge request on the same branch")

        # Try to get the local commit
        commit = self.gitm('show-ref', 'refs/heads/%s' % src).stdout.split()[0]
        # Do they exist on GitLab?
        try:
            srcb = repo.branches.get(src)
        except gitlab.GitlabGetError:
            srcb = None
            if self.question("Branch %s does not exist in your GitLab repo, shall I push?" % src):
                self.gitm('push', '-u', repo.remote, src, redirect=False)
            else:
                err("Aborting")
        if srcb and srcb.commit['id'] != commit:
            # Have we diverged? Then there are commits that are reachable from the GitLab branch but not local
            diverged = self.gitm('rev-list', srcb.commit['id'], '^' + commit)
            if diverged.stderr or diverged.stdout:
                if self.question("Branch %s has diverged from GitLab, shall I push and overwrite?" % src, default=False):
                    self.gitm('push', '--force', repo.remote, src, redirect=False)
                else:
                    err("Aborting")
            else:
                if self.question("Branch %s not up to date on GitLab, but can be fast forwarded, shall I push?" % src):
                    self.gitm('push', repo.remote, src, redirect=False)
                else:
                    err("Aborting")

        dstb = parent.branches.get(dst)
        if not dstb:
            err("Branch %s does not exist in %s/%s" % (dst, parent.namespace['full_path'], parent.path))

        # Do we have the dst locally?
        for remote in self.gitm('remote').stdout.strip().split("\n"):
            url = self.gitm('config', 'remote.%s.url' % remote).stdout.strip()
            if url in [parent.ssh_url_to_repo, parent.http_url_to_repo]:
                if not parent.visibility == 'public' and url != parent.ssh_url_to_repo:
                    err("You should configure %s/%s to fetch via ssh, it is a private repo" % (parent.namespace['full_path'], parent.path))
                self.gitm('fetch', remote, redirect=False)
                break
        else:
            err("You don't have %s/%s configured as a remote repository" % (parent.namespace['full_path'], parent.path))

        # How many commits?
        accept_empty_body = False
        commits = self.gitm('log', '--pretty=%H', '%s/%s..%s' % (remote, dst, src)).stdout.strip().split()
        commits.reverse()
        if not commits:
            err("Your branch has no commits yet")

        # 1 commit: title/body from commit
        if len(commits) == 1:
            title, body = self.gitm('log', '--pretty=%s\n%b', '%s^..%s' % (commits[0], commits[0])).stdout.split('\n', 1)
            title = title.strip()
            body = body.strip()
            accept_empty_body = not bool(body)

        # More commits: title from branchname (titlecased, s/-/ /g), body comments from shortlog
        else:
            title = src
            if '/' in title:
                title = title[title.rfind('/') + 1:]
            title = title.title().replace('-', ' ')
            body = ""

        extra = """Requesting a merge from %s/%s into %s/%s

Please enter a message to accompany your merge request. Lines starting
with '#' will be ignored, and an empty message aborts the request.""" % (repo.namespace['full_path'], src, parent.namespace['full_path'], dst)
        body += "\n\n" + self.gitm('shortlog', '%s/%s..%s' % (remote, dst, src)).stdout.strip()
        body += "\n\n" + self.gitm('diff', '--stat', '%s^..%s' % (commits[0], commits[-1])).stdout.strip()
        title, body = self.edit_msg(title, body, extra, 'MERGE_REQUEST_EDIT_MSG')
        if not body and not accept_empty_body:
            err("No merge request message specified")

        try:
            merge = parent.mergerequests.create({'project_id': repo.id, 'target_project_id': parent.id, 'source_branch': src, 'target_branch': dst, 'title': title, 'description': body})
            merge.save()
            print("merge request %d created %s" % (merge.iid, merge.web_url))
        except:
            filename = self.backup_message(title, body, 'merge-request-message-')
            err("Failed to create a merge request, the merge request text has been saved in %s" % filename)

    @command
    def mirror(self, opts):
        """[--ssh|--http] [<repo>]
           Mirror a repository, or all your repositories"""
        if opts['<repo>'] and opts['<repo>'] == '*':
            for repo in self.gl.Project():
                opts['<repo>'] = '%s/%s' % (repo.namespace['full_path'], name)
                self.mirror(opts)
            return
        repo = self.repository(opts)
        git_dir = repo.path + '.git'
        cur_dir = os.path.basename(os.path.abspath(os.getcwd()))
        if cur_dir != git_dir and not os.path.exists(git_dir):
            url = self.clone_url(repo, opts)
            self.gitm('clone', '--mirror', url, git_dir, redirect=False)
        else:
            if git_dir == cur_dir:
                git_dir = '.'
            # Update the current, mirrored repo
            if self.git('--git-dir', git_dir, 'config', 'core.bare').stdout.strip() != 'true' or \
               self.git('--git-dir', git_dir, 'config', 'remote.origin.mirror').stdout.strip() != 'true':
                   err("This is not a mirrored repository")
            self.gitm('--git-dir', git_dir, 'fetch', '-q', '--prune', 'origin', redirect=False)

        with open(os.path.join(git_dir, 'description'), 'w') as fd:
            fd.write(repo.description or "")

    @command
    def protect(self, opts):
        """<branch>
           Protect a branch against force-pushes"""
        repo = self.repository(opts)
        for branch in repo.branches.list():
            if branch.name == opts['<branch>']:
                branch.protect()
                break

    @command
    def protected(self, opts):
        """\nList protected branches"""
        repo = self.repository(opts)
        for branch in repo.branches.list():
            if branch.protected:
                print(branch.name)

    @command
    def public_keys(self, opts):
        """[<user>]
           Lists all keys for a user"""
        if opts['<user>']:
            user = self.gl.users.list(username=opts['<user>'][0])[0]
        else:
            user = self.me
        if not user:
            err("No such user")
        try:
            for key in user.keys.list():
                print("%s %s" % (key.key, key.title or ''))
        except gitlab.GitlabListError:
            # Permission denied, ignore
            pass

    @command
    def remove_member(self, opts):
        """<user>...
           Remove a user's membership"""
        repo = self.repository(opts)
        for member in repo.members.list():
            if member.username in opts['<user>']:
                member.delete()

    @command
    def repos(self, opts):
        """[--no-forks]
           List all your repos"""
        repos = self.gl.users.list(username=self.me.username)[0].projects.list()
        if not repos:
            return
        maxlen = max([len(x.name) for x in repos])
        fmt = u"%%-%ds %%s" % maxlen
        for repo in repos:
            color = [attr.normal]
            if repo.visibility == 'private':
                color.append(fgcolor.red)
            elif repo.visibility == 'internal':
                color.append(fgcolor.magenta)
            if hasattr(repo, 'forked_from_project'):
                if opts['--no-forks']:
                    continue
                color.append(attr.faint)
            name = repo.path
            if self.my_login != repo.namespace['full_path']:
                name = '%s/%s' % (repo.namespace['full_path'], name)
            desc = ' '.join((repo.description or '').splitlines())
            msg = wrap(fmt % (name, desc), *color)
            print(msg)

    @command
    def set_origin(self, opts, repo=None, remote='origin'):
        """[--ssh|--http] [--triangular [--upstream-branch=<branch>]]
           Set the remote 'origin' to gitlab.
           If this is a fork, set the remote 'upstream' to the parent"""
        if not repo:
            repo = self.repository(opts)
            # Is this mine? No? Do I have a clone?
            if repo.namespace['full_path'] != self.my_login:
                my_repo = self.gl.projects.get('%s/%s' % (self.my_login, repo.path))
                if my_repo:
                    repo = my_repo

        url = self.clone_url(repo, opts)
        if self.git('config', 'remote.%s.url' % remote).stdout.strip() != url:
            print("Pointing %s to %s" % (remote, url))
            self.gitm('config', 'remote.%s.url' % remote, url)
            self.gitm('config', 'remote.%s.gitlab-id' % remote, repo.id)
        self.gitm('config', '--replace-all', 'remote.%s.fetch' % remote, '+refs/heads/*:refs/remotes/%s/*' % remote)

        parent = self.parent_repo(repo)
        if parent:
            url = self.clone_url(parent, opts)
            if self.git('config', 'remote.upstream.url').stdout.strip() != url:
                print("Pointing upstream to %s" % url)
                self.gitm('config', 'remote.upstream.url', url)
                self.gitm('config', 'remote.upstream.gitlab-id', parent.id)
            self.gitm('config', 'remote.upstream.fetch', '+refs/heads/*:refs/remotes/upstream/*')

        if self.git('ls-remote', remote).stdout.strip():
            self.gitm('fetch', remote, redirect=False)
        if parent:
            self.gitm('fetch', 'upstream', redirect=False)

        if remote != 'origin':
            return

        self.set_tracking_branches(remote, upstream="upstream", triangular=opts['--triangular'], upstream_branch=opts['--upstream-branch'])

    @command
    def unprotect(self, opts):
        """<branch>
           Remove force-push protection from a branch"""
        repo = self.repository(opts)
        for branch in repo.branches.list():
            if branch.name == opts['<branch>']:
                branch.unprotect()
                break

    @command
    def whoami(self, opts):
        """\nDisplay GitLab user info"""
        opts['<user>'] = [self.me]
        self.whois(opts)

    @command
    def whois(self, opts):
        """<user>...
           Display GitLab user info"""
        for user in opts['<user>']:
            if not isinstance(user, (gitlab.v4.objects.User, gitlab.v4.objects.CurrentUser)):
                user_ = self.gl.users.list(username=user)
                if not user_:
                    print("No such user: %s" % user)
                    continue
                user = user_[0]
            print(wrap("%s (id %d)" % (user.name or user.username, user.id), attr.bright, attr.underline))
            print('Profile   %s' % user.web_url)
            if hasattr(user, 'email'):
                if user.email:
                    print('Email     %s' % user.email)
                if user.website_url:
                    print('Website   %s' % user.website_url)
                if user.twitter:
                    print('Twitter   %s' % user.twitter)
                if user.linkedin:
                    print('LinkedIn  %s' % user.linkedin)
                if user.bio:
                    if '\n' in user.bio:
                        bio = user.bio[:user.bio.index('\n')] + '...'
                    else:
                        bio = user.bio
                    print('Bio       %s' % user.bio)
            try:
                for pkey in user.keys.list():
                    algo, key = pkey.key.split()[:2]
                    algo = algo[4:].upper()
                    if pkey.title:
                        print("%s key%s...%s (%s)" % (algo, ' ' * (6 - len(algo)), key[-10:], pkey.title))
                    else:
                        print("%s key%s...%s" % (algo, ' ' * (6 - len(algo)), key[-10:]))

            except gitlab.GitlabListError:
                # Permission denied, ignore
                pass



from setuptools import setup

setup(name='git-spindle',
    version="4.0.1",
    description='Git subcommands for integrating with central services like github, gitlab and bitbucket',
    author='Dennis Kaarsemaker',
    author_email='dennis@kaarsemaker.net',
    url='http://github.com/seveas/git-spindle',
    packages=['gitspindle'],
    package_dir={'': 'lib'},
    entry_points={
        'console_scripts':[
            'git-hub=gitspindle:hub',
            'git-lab=gitspindle:lab',
            'git-bucket=gitspindle:bucket',
            'git-bb=gitspindle:bb',
        ]
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Topic :: Software Development",
        "Topic :: Software Development :: Version Control"
    ],
    install_requires=["github3.py>=1.3.0", "python-gitlab>=2.5.0", "whelk>=3.0", "docopt>=0.6.2"],
)
